from .common import InfoExtractor
from ..utils import unified_strdate


class CozyTVIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?cozy\.tv/(?P<uploader>[^/]+)/replays/(?P<id>[^/$#&?]+)'

    _TESTS = [{
        'url': 'https://cozy.tv/beardson/replays/2021-11-19_1',
        'info_dict': {
            'id': 'beardson-2021-11-19_1',
            'ext': 'mp4',
            'title': 'pokemon pt2',
            'uploader': 'beardson',
            'upload_date': '20211119',
            'was_live': True,
            'duration': 7981,
        },
        'params': {'skip_download': True},
    }]

    def _real_extract(self, url):
        uploader, date = self._match_valid_url(url).groups()
        video_id = f'{uploader}-{date}'
        data_json = self._download_json(f'https://api.cozy.tv/cache/{uploader}/replay/{date}', video_id)
        formats, subtitles = self._extract_m3u8_formats_and_subtitles(
            f'https://cozycdn.foxtrotstream.xyz/replays/{uploader}/{date}/index.m3u8', video_id, ext='mp4')
        return {
            'id': video_id,
            'title': data_json.get('title'),
            'uploader': data_json.get('user') or uploader,
            'upload_date': unified_strdate(data_json.get('date')),
            'was_live': True,
            'duration': data_json.get('duration'),
            'formats': formats,
            'subtitles': subtitles,
        }



import re

from .common import InfoExtractor


class SoundgasmIE(InfoExtractor):
    IE_NAME = 'soundgasm'
    _VALID_URL = r'https?://(?:www\.)?soundgasm\.net/u/(?P<user>[0-9a-zA-Z_-]+)/(?P<display_id>[0-9a-zA-Z_-]+)'
    _TEST = {
        'url': 'http://soundgasm.net/u/ytdl/Piano-sample',
        'md5': '010082a2c802c5275bb00030743e75ad',
        'info_dict': {
            'id': '88abd86ea000cafe98f96321b23cc1206cbcbcc9',
            'ext': 'm4a',
            'title': 'Piano sample',
            'description': 'Royalty Free Sample Music',
            'uploader': 'ytdl',
        },
    }

    def _real_extract(self, url):
        mobj = self._match_valid_url(url)
        display_id = mobj.group('display_id')

        webpage = self._download_webpage(url, display_id)

        audio_url = self._html_search_regex(
            r'(?s)m4a\s*:\s*(["\'])(?P<url>(?:(?!\1).)+)\1', webpage,
            'audio URL', group='url')

        title = self._search_regex(
            r'<div[^>]+\bclass=["\']jp-title[^>]+>([^<]+)',
            webpage, 'title', default=display_id)

        description = self._html_search_regex(
            (r'(?s)<div[^>]+\bclass=["\']jp-description[^>]+>(.+?)</div>',
             r'(?s)<li>Description:\s(.*?)<\/li>'),
            webpage, 'description', fatal=False)

        audio_id = self._search_regex(
            r'/([^/]+)\.m4a', audio_url, 'audio id', default=display_id)

        return {
            'id': audio_id,
            'display_id': display_id,
            'url': audio_url,
            'vcodec': 'none',
            'title': title,
            'description': description,
            'uploader': mobj.group('user'),
        }


class SoundgasmProfileIE(InfoExtractor):
    IE_NAME = 'soundgasm:profile'
    _VALID_URL = r'https?://(?:www\.)?soundgasm\.net/u/(?P<id>[^/]+)/?(?:\#.*)?$'
    _TEST = {
        'url': 'http://soundgasm.net/u/ytdl',
        'info_dict': {
            'id': 'ytdl',
        },
        'playlist_count': 1,
    }

    def _real_extract(self, url):
        profile_id = self._match_id(url)

        webpage = self._download_webpage(url, profile_id)

        entries = [
            self.url_result(audio_url, 'Soundgasm')
            for audio_url in re.findall(rf'href="([^"]+/u/{profile_id}/[^"]+)', webpage)]

        return self.playlist_result(entries, profile_id)



import getpass
import json
import re
import time
import urllib.parse
import uuid
import xml.etree.ElementTree as etree

from .common import InfoExtractor
from ..networking.exceptions import HTTPError
from ..utils import (
    NO_DEFAULT,
    ExtractorError,
    parse_qs,
    unescapeHTML,
    unified_timestamp,
    urlencode_postdata,
)

MSO_INFO = {
    'DTV': {
        'name': 'DIRECTV',
        'username_field': 'username',
        'password_field': 'password',
    },
    'ATT': {
        'name': 'AT&T U-verse',
        'username_field': 'userid',
        'password_field': 'password',
    },
    'ATTOTT': {
        'name': 'DIRECTV NOW',
        'username_field': 'email',
        'password_field': 'loginpassword',
    },
    'RCN': {
        'name': 'RCN',
        'username_field': 'username',
        'password_field': 'password',
    },
    'Rogers': {
        'name': 'Rogers',
        'username_field': 'UserName',
        'password_field': 'UserPassword',
    },
    'Comcast_SSO': {
        'name': 'Comcast XFINITY',
        'username_field': 'user',
        'password_field': 'passwd',
        'login_hostname': 'login.xfinity.com',
    },
    'TWC': {
        'name': 'Time Warner Cable | Spectrum',
        'username_field': 'Ecom_User_ID',
        'password_field': 'Ecom_Password',
    },
    'Brighthouse': {
        'name': 'Bright House Networks | Spectrum',
        'username_field': 'j_username',
        'password_field': 'j_password',
    },
    'Charter_Direct': {
        'name': 'Charter Spectrum',
        'username_field': 'IDToken1',
        'password_field': 'IDToken2',
    },
    'Spectrum': {
        'name': 'Spectrum',
        'username_field': 'IDToken1',
        'password_field': 'IDToken2',
    },
    'Philo': {
        'name': 'Philo',
        'username_field': 'ident',
    },
    'Verizon': {
        'name': 'Verizon FiOS',
        'username_field': 'IDToken1',
        'password_field': 'IDToken2',
        'login_hostname': 'ssoauth.verizon.com',
    },
    'Fubo': {
        'name': 'Fubo',
        'username_field': 'username',
        'password_field': 'password',
    },
    'Cablevision': {
        'name': 'Optimum/Cablevision',
        'username_field': 'j_username',
        'password_field': 'j_password',
    },
    'thr030': {
        'name': '3 Rivers Communications',
    },
    'com140': {
        'name': 'Access Montana',
    },
    'acecommunications': {
        'name': 'AcenTek',
    },
    'acm010': {
        'name': 'Acme Communications',
    },
    'ada020': {
        'name': 'Adams Cable Service',
    },
    'alb020': {
        'name': 'Albany Mutual Telephone',
    },
    'algona': {
        'name': 'Algona Municipal Utilities',
    },
    'allwest': {
        'name': 'All West Communications',
    },
    'all025': {
        'name': 'Allen\'s Communications',
    },
    'spl010': {
        'name': 'Alliance Communications',
    },
    'all070': {
        'name': 'ALLO Communications',
    },
    'alpine': {
        'name': 'Alpine Communications',
    },
    'hun015': {
        'name': 'American Broadband',
    },
    'nwc010': {
        'name': 'American Broadband Missouri',
    },
    'com130-02': {
        'name': 'American Community Networks',
    },
    'com130-01': {
        'name': 'American Warrior Networks',
    },
    'tom020': {
        'name': 'Amherst Telephone/Tomorrow Valley',
    },
    'tvc020': {
        'name': 'Andycable',
    },
    'arkwest': {
        'name': 'Arkwest Communications',
    },
    'art030': {
        'name': 'Arthur Mutual Telephone Company',
    },
    'arvig': {
        'name': 'Arvig',
    },
    'nttcash010': {
        'name': 'Ashland Home Net',
    },
    'astound': {
        'name': 'Astound (now Wave)',
    },
    'dix030': {
        'name': 'ATC Broadband',
    },
    'ara010': {
        'name': 'ATC Communications',
    },
    'she030-02': {
        'name': 'Ayersville Communications',
    },
    'baldwin': {
        'name': 'Baldwin Lightstream',
    },
    'bal040': {
        'name': 'Ballard TV',
    },
    'cit025': {
        'name': 'Bardstown Cable TV',
    },
    'bay030': {
        'name': 'Bay Country Communications',
    },
    'tel095': {
        'name': 'Beaver Creek Cooperative Telephone',
    },
    'bea020': {
        'name': 'Beaver Valley Cable',
    },
    'bee010': {
        'name': 'Bee Line Cable',
    },
    'wir030': {
        'name': 'Beehive Broadband',
    },
    'bra020': {
        'name': 'BELD',
    },
    'bel020': {
        'name': 'Bellevue Municipal Cable',
    },
    'vol040-01': {
        'name': 'Ben Lomand Connect / BLTV',
    },
    'bev010': {
        'name': 'BEVCOMM',
    },
    'big020': {
        'name': 'Big Sandy Broadband',
    },
    'ble020': {
        'name': 'Bledsoe Telephone Cooperative',
    },
    'bvt010': {
        'name': 'Blue Valley Tele-Communications',
    },
    'bra050': {
        'name': 'Brandenburg Telephone Co.',
    },
    'bte010': {
        'name': 'Bristol Tennessee Essential Services',
    },
    'annearundel': {
        'name': 'Broadstripe',
    },
    'btc010': {
        'name': 'BTC Communications',
    },
    'btc040': {
        'name': 'BTC Vision - Nahunta',
    },
    'bul010': {
        'name': 'Bulloch Telephone Cooperative',
    },
    'but010': {
        'name': 'Butler-Bremer Communications',
    },
    'tel160-csp': {
        'name': 'C Spire SNAP',
    },
    'csicable': {
        'name': 'Cable Services Inc.',
    },
    'cableamerica': {
        'name': 'CableAmerica',
    },
    'cab038': {
        'name': 'CableSouth Media 3',
    },
    'weh010-camtel': {
        'name': 'Cam-Tel Company',
    },
    'car030': {
        'name': 'Cameron Communications',
    },
    'canbytel': {
        'name': 'Canby Telcom',
    },
    'crt020': {
        'name': 'CapRock Tv',
    },
    'car050': {
        'name': 'Carnegie Cable',
    },
    'cas': {
        'name': 'CAS Cable',
    },
    'casscomm': {
        'name': 'CASSCOMM',
    },
    'mid180-02': {
        'name': 'Catalina Broadband Solutions',
    },
    'cccomm': {
        'name': 'CC Communications',
    },
    'nttccde010': {
        'name': 'CDE Lightband',
    },
    'cfunet': {
        'name': 'Cedar Falls Utilities',
    },
    'dem010-01': {
        'name': 'Celect-Bloomer Telephone Area',
    },
    'dem010-02': {
        'name': 'Celect-Bruce Telephone Area',
    },
    'dem010-03': {
        'name': 'Celect-Citizens Connected Area',
    },
    'dem010-04': {
        'name': 'Celect-Elmwood/Spring Valley Area',
    },
    'dem010-06': {
        'name': 'Celect-Mosaic Telecom',
    },
    'dem010-05': {
        'name': 'Celect-West WI Telephone Area',
    },
    'net010-02': {
        'name': 'Cellcom/Nsight Telservices',
    },
    'cen100': {
        'name': 'CentraCom',
    },
    'nttccst010': {
        'name': 'Central Scott / CSTV',
    },
    'cha035': {
        'name': 'Chaparral CableVision',
    },
    'cha050': {
        'name': 'Chariton Valley Communication Corporation, Inc.',
    },
    'cha060': {
        'name': 'Chatmoss Cablevision',
    },
    'nttcche010': {
        'name': 'Cherokee Communications',
    },
    'che050': {
        'name': 'Chesapeake Bay Communications',
    },
    'cimtel': {
        'name': 'Cim-Tel Cable, LLC.',
    },
    'cit180': {
        'name': 'Citizens Cablevision - Floyd, VA',
    },
    'cit210': {
        'name': 'Citizens Cablevision, Inc.',
    },
    'cit040': {
        'name': 'Citizens Fiber',
    },
    'cit250': {
        'name': 'Citizens Mutual',
    },
    'war040': {
        'name': 'Citizens Telephone Corporation',
    },
    'wat025': {
        'name': 'City Of Monroe',
    },
    'wadsworth': {
        'name': 'CityLink',
    },
    'nor100': {
        'name': 'CL Tel',
    },
    'cla010': {
        'name': 'Clarence Telephone and Cedar Communications',
    },
    'ser060': {
        'name': 'Clear Choice Communications',
    },
    'tac020': {
        'name': 'Click! Cable TV',
    },
    'war020': {
        'name': 'CLICK1.NET',
    },
    'cml010': {
        'name': 'CML Telephone Cooperative Association',
    },
    'cns': {
        'name': 'CNS',
    },
    'com160': {
        'name': 'Co-Mo Connect',
    },
    'coa020': {
        'name': 'Coast Communications',
    },
    'coa030': {
        'name': 'Coaxial Cable TV',
    },
    'mid055': {
        'name': 'Cobalt TV (Mid-State Community TV)',
    },
    'col070': {
        'name': 'Columbia Power & Water Systems',
    },
    'col080': {
        'name': 'Columbus Telephone',
    },
    'nor105': {
        'name': 'Communications 1 Cablevision, Inc.',
    },
    'com150': {
        'name': 'Community Cable & Broadband',
    },
    'com020': {
        'name': 'Community Communications Company',
    },
    'coy010': {
        'name': 'commZoom',
    },
    'com025': {
        'name': 'Complete Communication Services',
    },
    'cat020': {
        'name': 'Comporium',
    },
    'com071': {
        'name': 'ComSouth Telesys',
    },
    'consolidatedcable': {
        'name': 'Consolidated',
    },
    'conwaycorp': {
        'name': 'Conway Corporation',
    },
    'coo050': {
        'name': 'Coon Valley Telecommunications Inc',
    },
    'coo080': {
        'name': 'Cooperative Telephone Company',
    },
    'cpt010': {
        'name': 'CP-TEL',
    },
    'cra010': {
        'name': 'Craw-Kan Telephone',
    },
    'crestview': {
        'name': 'Crestview Cable Communications',
    },
    'cross': {
        'name': 'Cross TV',
    },
    'cro030': {
        'name': 'Crosslake Communications',
    },
    'ctc040': {
        'name': 'CTC - Brainerd MN',
    },
    'phe030': {
        'name': 'CTV-Beam - East Alabama',
    },
    'cun010': {
        'name': 'Cunningham Telephone & Cable',
    },
    'dpc010': {
        'name': 'D & P Communications',
    },
    'dak030': {
        'name': 'Dakota Central Telecommunications',
    },
    'nttcdel010': {
        'name': 'Delcambre Telephone LLC',
    },
    'tel160-del': {
        'name': 'Delta Telephone Company',
    },
    'sal040': {
        'name': 'DiamondNet',
    },
    'ind060-dc': {
        'name': 'Direct Communications',
    },
    'doy010': {
        'name': 'Doylestown Cable TV',
    },
    'dic010': {
        'name': 'DRN',
    },
    'dtc020': {
        'name': 'DTC',
    },
    'dtc010': {
        'name': 'DTC Cable (Delhi)',
    },
    'dum010': {
        'name': 'Dumont Telephone Company',
    },
    'dun010': {
        'name': 'Dunkerton Telephone Cooperative',
    },
    'cci010': {
        'name': 'Duo County Telecom',
    },
    'eagle': {
        'name': 'Eagle Communications',
    },
    'weh010-east': {
        'name': 'East Arkansas Cable TV',
    },
    'eatel': {
        'name': 'EATEL Video, LLC',
    },
    'ell010': {
        'name': 'ECTA',
    },
    'emerytelcom': {
        'name': 'Emery Telcom Video LLC',
    },
    'nor200': {
        'name': 'Empire Access',
    },
    'endeavor': {
        'name': 'Endeavor Communications',
    },
    'sun045': {
        'name': 'Enhanced Telecommunications Corporation',
    },
    'mid030': {
        'name': 'enTouch',
    },
    'epb020': {
        'name': 'EPB Smartnet',
    },
    'jea010': {
        'name': 'EPlus Broadband',
    },
    'com065': {
        'name': 'ETC',
    },
    'ete010': {
        'name': 'Etex Communications',
    },
    'fbc-tele': {
        'name': 'F&B Communications',
    },
    'fal010': {
        'name': 'Falcon Broadband',
    },
    'fam010': {
        'name': 'FamilyView CableVision',
    },
    'far020': {
        'name': 'Farmers Mutual Telephone Company',
    },
    'fay010': {
        'name': 'Fayetteville Public Utilities',
    },
    'sal060': {
        'name': 'fibrant',
    },
    'fid010': {
        'name': 'Fidelity Communications',
    },
    'for030': {
        'name': 'FJ Communications',
    },
    'fli020': {
        'name': 'Flint River Communications',
    },
    'far030': {
        'name': 'FMT - Jesup',
    },
    'foo010': {
        'name': 'Foothills Communications',
    },
    'for080': {
        'name': 'Forsyth CableNet',
    },
    'fbcomm': {
        'name': 'Frankfort Plant Board',
    },
    'tel160-fra': {
        'name': 'Franklin Telephone Company',
    },
    'nttcftc010': {
        'name': 'FTC',
    },
    'fullchannel': {
        'name': 'Full Channel, Inc.',
    },
    'gar040': {
        'name': 'Gardonville Cooperative Telephone Association',
    },
    'gbt010': {
        'name': 'GBT Communications, Inc.',
    },
    'tec010': {
        'name': 'Genuine Telecom',
    },
    'clr010': {
        'name': 'Giant Communications',
    },
    'gla010': {
        'name': 'Glasgow EPB',
    },
    'gle010': {
        'name': 'Glenwood Telecommunications',
    },
    'gra060': {
        'name': 'GLW Broadband Inc.',
    },
    'goldenwest': {
        'name': 'Golden West Cablevision',
    },
    'vis030': {
        'name': 'Grantsburg Telcom',
    },
    'gpcom': {
        'name': 'Great Plains Communications',
    },
    'gri010': {
        'name': 'Gridley Cable Inc',
    },
    'hbc010': {
        'name': 'H&B Cable Services',
    },
    'hae010': {
        'name': 'Haefele TV Inc.',
    },
    'htc010': {
        'name': 'Halstad Telephone Company',
    },
    'har005': {
        'name': 'Harlan Municipal Utilities',
    },
    'har020': {
        'name': 'Hart Communications',
    },
    'ced010': {
        'name': 'Hartelco TV',
    },
    'hea040': {
        'name': 'Heart of Iowa Communications Cooperative',
    },
    'htc020': {
        'name': 'Hickory Telephone Company',
    },
    'nttchig010': {
        'name': 'Highland Communication Services',
    },
    'hig030': {
        'name': 'Highland Media',
    },
    'spc010': {
        'name': 'Hilliary Communications',
    },
    'hin020': {
        'name': 'Hinton CATV Co.',
    },
    'hometel': {
        'name': 'HomeTel Entertainment, Inc.',
    },
    'hoodcanal': {
        'name': 'Hood Canal Communications',
    },
    'weh010-hope': {
        'name': 'Hope - Prescott Cable TV',
    },
    'horizoncable': {
        'name': 'Horizon Cable TV, Inc.',
    },
    'hor040': {
        'name': 'Horizon Chillicothe Telephone',
    },
    'htc030': {
        'name': 'HTC Communications Co. - IL',
    },
    'htccomm': {
        'name': 'HTC Communications, Inc. - IA',
    },
    'wal005': {
        'name': 'Huxley Communications',
    },
    'imon': {
        'name': 'ImOn Communications',
    },
    'ind040': {
        'name': 'Independence Telecommunications',
    },
    'rrc010': {
        'name': 'Inland Networks',
    },
    'stc020': {
        'name': 'Innovative Cable TV St Croix',
    },
    'car100': {
        'name': 'Innovative Cable TV St Thomas-St John',
    },
    'icc010': {
        'name': 'Inside Connect Cable',
    },
    'int100': {
        'name': 'Integra Telecom',
    },
    'int050': {
        'name': 'Interstate Telecommunications Coop',
    },
    'irv010': {
        'name': 'Irvine Cable',
    },
    'k2c010': {
        'name': 'K2 Communications',
    },
    'kal010': {
        'name': 'Kalida Telephone Company, Inc.',
    },
    'kal030': {
        'name': 'Kalona Cooperative Telephone Company',
    },
    'kmt010': {
        'name': 'KMTelecom',
    },
    'kpu010': {
        'name': 'KPU Telecommunications',
    },
    'kuh010': {
        'name': 'Kuhn Communications, Inc.',
    },
    'lak130': {
        'name': 'Lakeland Communications',
    },
    'lan010': {
        'name': 'Langco',
    },
    'lau020': {
        'name': 'Laurel Highland Total Communications, Inc.',
    },
    'leh010': {
        'name': 'Lehigh Valley Cooperative Telephone',
    },
    'bra010': {
        'name': 'Limestone Cable/Bracken Cable',
    },
    'loc020': {
        'name': 'LISCO',
    },
    'lit020': {
        'name': 'Litestream',
    },
    'tel140': {
        'name': 'LivCom',
    },
    'loc010': {
        'name': 'LocalTel Communications',
    },
    'weh010-longview': {
        'name': 'Longview - Kilgore Cable TV',
    },
    'lon030': {
        'name': 'Lonsdale Video Ventures, LLC',
    },
    'lns010': {
        'name': 'Lost Nation-Elwood Telephone Co.',
    },
    'nttclpc010': {
        'name': 'LPC Connect',
    },
    'lumos': {
        'name': 'Lumos Networks',
    },
    'madison': {
        'name': 'Madison Communications',
    },
    'mad030': {
        'name': 'Madison County Cable Inc.',
    },
    'nttcmah010': {
        'name': 'Mahaska Communication Group',
    },
    'mar010': {
        'name': 'Marne & Elk Horn Telephone Company',
    },
    'mcc040': {
        'name': 'McClure Telephone Co.',
    },
    'mctv': {
        'name': 'MCTV',
    },
    'merrimac': {
        'name': 'Merrimac Communications Ltd.',
    },
    'metronet': {
        'name': 'Metronet',
    },
    'mhtc': {
        'name': 'MHTC',
    },
    'midhudson': {
        'name': 'Mid-Hudson Cable',
    },
    'midrivers': {
        'name': 'Mid-Rivers Communications',
    },
    'mid045': {
        'name': 'Midstate Communications',
    },
    'mil080': {
        'name': 'Milford Communications',
    },
    'min030': {
        'name': 'MINET',
    },
    'nttcmin010': {
        'name': 'Minford TV',
    },
    'san040-02': {
        'name': 'Mitchell Telecom',
    },
    'mlg010': {
        'name': 'MLGC',
    },
    'mon060': {
        'name': 'Mon-Cre TVE',
    },
    'mou110': {
        'name': 'Mountain Telephone',
    },
    'mou050': {
        'name': 'Mountain Village Cable',
    },
    'mtacomm': {
        'name': 'MTA Communications, LLC',
    },
    'mtc010': {
        'name': 'MTC Cable',
    },
    'med040': {
        'name': 'MTC Technologies',
    },
    'man060': {
        'name': 'MTCC',
    },
    'mtc030': {
        'name': 'MTCO Communications',
    },
    'mul050': {
        'name': 'Mulberry Telecommunications',
    },
    'mur010': {
        'name': 'Murray Electric System',
    },
    'musfiber': {
        'name': 'MUS FiberNET',
    },
    'mpw': {
        'name': 'Muscatine Power & Water',
    },
    'nttcsli010': {
        'name': 'myEVTV.com',
    },
    'nor115': {
        'name': 'NCC',
    },
    'nor260': {
        'name': 'NDTC',
    },
    'nctc': {
        'name': 'Nebraska Central Telecom, Inc.',
    },
    'nel020': {
        'name': 'Nelsonville TV Cable',
    },
    'nem010': {
        'name': 'Nemont',
    },
    'new075': {
        'name': 'New Hope Telephone Cooperative',
    },
    'nor240': {
        'name': 'NICP',
    },
    'cic010': {
        'name': 'NineStar Connect',
    },
    'nktelco': {
        'name': 'NKTelco',
    },
    'nortex': {
        'name': 'Nortex Communications',
    },
    'nor140': {
        'name': 'North Central Telephone Cooperative',
    },
    'nor030': {
        'name': 'Northland Communications',
    },
    'nor075': {
        'name': 'Northwest Communications',
    },
    'nor125': {
        'name': 'Norwood Light Broadband',
    },
    'net010': {
        'name': 'Nsight Telservices',
    },
    'dur010': {
        'name': 'Ntec',
    },
    'nts010': {
        'name': 'NTS Communications',
    },
    'new045': {
        'name': 'NU-Telecom',
    },
    'nulink': {
        'name': 'NuLink',
    },
    'jam030': {
        'name': 'NVC',
    },
    'far035': {
        'name': 'OmniTel Communications',
    },
    'onesource': {
        'name': 'OneSource Communications',
    },
    'cit230': {
        'name': 'Opelika Power Services',
    },
    'daltonutilities': {
        'name': 'OptiLink',
    },
    'mid140': {
        'name': 'OPTURA',
    },
    'ote010': {
        'name': 'OTEC Communication Company',
    },
    'cci020': {
        'name': 'Packerland Broadband',
    },
    'pan010': {
        'name': 'Panora Telco/Guthrie Center Communications',
    },
    'otter': {
        'name': 'Park Region Telephone & Otter Tail Telcom',
    },
    'mid050': {
        'name': 'Partner Communications Cooperative',
    },
    'fib010': {
        'name': 'Pathway',
    },
    'paulbunyan': {
        'name': 'Paul Bunyan Communications',
    },
    'pem020': {
        'name': 'Pembroke Telephone Company',
    },
    'mck010': {
        'name': 'Peoples Rural Telephone Cooperative',
    },
    'pul010': {
        'name': 'PES Energize',
    },
    'phi010': {
        'name': 'Philippi Communications System',
    },
    'phonoscope': {
        'name': 'Phonoscope Cable',
    },
    'pin070': {
        'name': 'Pine Belt Communications, Inc.',
    },
    'weh010-pine': {
        'name': 'Pine Bluff Cable TV',
    },
    'pin060': {
        'name': 'Pineland Telephone Cooperative',
    },
    'cam010': {
        'name': 'Pinpoint Communications',
    },
    'pio060': {
        'name': 'Pioneer Broadband',
    },
    'pioncomm': {
        'name': 'Pioneer Communications',
    },
    'pioneer': {
        'name': 'Pioneer DTV',
    },
    'pla020': {
        'name': 'Plant TiftNet, Inc.',
    },
    'par010': {
        'name': 'PLWC',
    },
    'pro035': {
        'name': 'PMT',
    },
    'vik011': {
        'name': 'Polar Cablevision',
    },
    'pottawatomie': {
        'name': 'Pottawatomie Telephone Co.',
    },
    'premiercomm': {
        'name': 'Premier Communications',
    },
    'psc010': {
        'name': 'PSC',
    },
    'pan020': {
        'name': 'PTCI',
    },
    'qco010': {
        'name': 'QCOL',
    },
    'qua010': {
        'name': 'Quality Cablevision',
    },
    'rad010': {
        'name': 'Radcliffe Telephone Company',
    },
    'car040': {
        'name': 'Rainbow Communications',
    },
    'rai030': {
        'name': 'Rainier Connect',
    },
    'ral010': {
        'name': 'Ralls Technologies',
    },
    'rct010': {
        'name': 'RC Technologies',
    },
    'red040': {
        'name': 'Red River Communications',
    },
    'ree010': {
        'name': 'Reedsburg Utility Commission',
    },
    'mol010': {
        'name': 'Reliance Connects- Oregon',
    },
    'res020': {
        'name': 'Reserve Telecommunications',
    },
    'weh010-resort': {
        'name': 'Resort TV Cable',
    },
    'rld010': {
        'name': 'Richland Grant Telephone Cooperative, Inc.',
    },
    'riv030': {
        'name': 'River Valley Telecommunications Coop',
    },
    'rockportcable': {
        'name': 'Rock Port Cablevision',
    },
    'rsf010': {
        'name': 'RS Fiber',
    },
    'rtc': {
        'name': 'RTC Communication Corp',
    },
    'res040': {
        'name': 'RTC-Reservation Telephone Coop.',
    },
    'rte010': {
        'name': 'RTEC Communications',
    },
    'stc010': {
        'name': 'S&T',
    },
    'san020': {
        'name': 'San Bruno Cable TV',
    },
    'san040-01': {
        'name': 'Santel',
    },
    'sav010': {
        'name': 'SCI Broadband-Savage Communications Inc.',
    },
    'sco050': {
        'name': 'Scottsboro Electric Power Board',
    },
    'scr010': {
        'name': 'Scranton Telephone Company',
    },
    'selco': {
        'name': 'SELCO',
    },
    'she010': {
        'name': 'Shentel',
    },
    'she030': {
        'name': 'Sherwood Mutual Telephone Association, Inc.',
    },
    'ind060-ssc': {
        'name': 'Silver Star Communications',
    },
    'sjoberg': {
        'name': 'Sjoberg\'s Inc.',
    },
    'sou025': {
        'name': 'SKT',
    },
    'sky050': {
        'name': 'SkyBest TV',
    },
    'nttcsmi010': {
        'name': 'Smithville Communications',
    },
    'woo010': {
        'name': 'Solarus',
    },
    'sou075': {
        'name': 'South Central Rural Telephone Cooperative',
    },
    'sou065': {
        'name': 'South Holt Cablevision, Inc.',
    },
    'sou035': {
        'name': 'South Slope Cooperative Communications',
    },
    'spa020': {
        'name': 'Spanish Fork Community Network',
    },
    'spe010': {
        'name': 'Spencer Municipal Utilities',
    },
    'spi005': {
        'name': 'Spillway Communications, Inc.',
    },
    'srt010': {
        'name': 'SRT',
    },
    'cccsmc010': {
        'name': 'St. Maarten Cable TV',
    },
    'sta025': {
        'name': 'Star Communications',
    },
    'sco020': {
        'name': 'STE',
    },
    'uin010': {
        'name': 'STRATA Networks',
    },
    'sum010': {
        'name': 'Sumner Cable TV',
    },
    'pie010': {
        'name': 'Surry TV/PCSI TV',
    },
    'swa010': {
        'name': 'Swayzee Communications',
    },
    'sweetwater': {
        'name': 'Sweetwater Cable Television Co',
    },
    'weh010-talequah': {
        'name': 'Tahlequah Cable TV',
    },
    'tct': {
        'name': 'TCT',
    },
    'tel050': {
        'name': 'Tele-Media Company',
    },
    'com050': {
        'name': 'The Community Agency',
    },
    'thr020': {
        'name': 'Three River',
    },
    'cab140': {
        'name': 'Town & Country Technologies',
    },
    'tra010': {
        'name': 'Trans-Video',
    },
    'tre010': {
        'name': 'Trenton TV Cable Company',
    },
    'tcc': {
        'name': 'Tri County Communications Cooperative',
    },
    'tri025': {
        'name': 'TriCounty Telecom',
    },
    'tri110': {
        'name': 'TrioTel Communications, Inc.',
    },
    'tro010': {
        'name': 'Troy Cablevision, Inc.',
    },
    'tsc': {
        'name': 'TSC',
    },
    'cit220': {
        'name': 'Tullahoma Utilities Board',
    },
    'tvc030': {
        'name': 'TV Cable of Rensselaer',
    },
    'tvc015': {
        'name': 'TVC Cable',
    },
    'cab180': {
        'name': 'TVision',
    },
    'twi040': {
        'name': 'Twin Lakes',
    },
    'tvtinc': {
        'name': 'Twin Valley',
    },
    'uis010': {
        'name': 'Union Telephone Company',
    },
    'uni110': {
        'name': 'United Communications - TN',
    },
    'uni120': {
        'name': 'United Services',
    },
    'uss020': {
        'name': 'US Sonet',
    },
    'cab060': {
        'name': 'USA Communications',
    },
    'she005': {
        'name': 'USA Communications/Shellsburg, IA',
    },
    'val040': {
        'name': 'Valley TeleCom Group',
    },
    'val025': {
        'name': 'Valley Telecommunications',
    },
    'val030': {
        'name': 'Valparaiso Broadband',
    },
    'cla050': {
        'name': 'Vast Broadband',
    },
    'sul015': {
        'name': 'Venture Communications Cooperative, Inc.',
    },
    'ver025': {
        'name': 'Vernon Communications Co-op',
    },
    'weh010-vicksburg': {
        'name': 'Vicksburg Video',
    },
    'vis070': {
        'name': 'Vision Communications',
    },
    'volcanotel': {
        'name': 'Volcano Vision, Inc.',
    },
    'vol040-02': {
        'name': 'VolFirst / BLTV',
    },
    'ver070': {
        'name': 'VTel',
    },
    'nttcvtx010': {
        'name': 'VTX1',
    },
    'bci010-02': {
        'name': 'Vyve Broadband',
    },
    'wab020': {
        'name': 'Wabash Mutual Telephone',
    },
    'waitsfield': {
        'name': 'Waitsfield Cable',
    },
    'wal010': {
        'name': 'Walnut Communications',
    },
    'wavebroadband': {
        'name': 'Wave',
    },
    'wav030': {
        'name': 'Waverly Communications Utility',
    },
    'wbi010': {
        'name': 'WBI',
    },
    'web020': {
        'name': 'Webster-Calhoun Cooperative Telephone Association',
    },
    'wes005': {
        'name': 'West Alabama TV Cable',
    },
    'carolinata': {
        'name': 'West Carolina Communications',
    },
    'wct010': {
        'name': 'West Central Telephone Association',
    },
    'wes110': {
        'name': 'West River Cooperative Telephone Company',
    },
    'ani030': {
        'name': 'WesTel Systems',
    },
    'westianet': {
        'name': 'Western Iowa Networks',
    },
    'nttcwhi010': {
        'name': 'Whidbey Telecom',
    },
    'weh010-white': {
        'name': 'White County Cable TV',
    },
    'wes130': {
        'name': 'Wiatel',
    },
    'wik010': {
        'name': 'Wiktel',
    },
    'wil070': {
        'name': 'Wilkes Communications, Inc./RiverStreet Networks',
    },
    'wil015': {
        'name': 'Wilson Communications',
    },
    'win010': {
        'name': 'Windomnet/SMBS',
    },
    'win090': {
        'name': 'Windstream Cable TV',
    },
    'wcta': {
        'name': 'Winnebago Cooperative Telecom Association',
    },
    'wtc010': {
        'name': 'WTC',
    },
    'wil040': {
        'name': 'WTC Communications, Inc.',
    },
    'wya010': {
        'name': 'Wyandotte Cable',
    },
    'hin020-02': {
        'name': 'X-Stream Services',
    },
    'xit010': {
        'name': 'XIT Communications',
    },
    'yel010': {
        'name': 'Yelcot Communications',
    },
    'mid180-01': {
        'name': 'yondoo',
    },
    'cou060': {
        'name': 'Zito Media',
    },
    'slingtv': {
        'name': 'Sling TV',
        'username_field': 'username',
        'password_field': 'password',
        'login_hostname': 'identity.sling.com',
    },
    'Suddenlink': {
        'name': 'Suddenlink',
        'username_field': 'username',
        'password_field': 'password',
    },
    'AlticeOne': {
        'name': 'Optimum TV',
        'username_field': 'j_username',
        'password_field': 'j_password',
    },
}


class AdobePassIE(InfoExtractor):  # XXX: Conventionally, base classes should end with BaseIE/InfoExtractor
    _SERVICE_PROVIDER_TEMPLATE = 'https://sp.auth.adobe.com/adobe-services/%s'
    _USER_AGENT = 'Mozilla/5.0 (X11; Linux i686; rv:47.0) Gecko/20100101 Firefox/47.0'
    _MVPD_CACHE = 'ap-mvpd'

    _DOWNLOADING_LOGIN_PAGE = 'Downloading Provider Login Page'

    def _download_webpage_handle(self, *args, **kwargs):
        headers = self.geo_verification_headers()
        headers.update(kwargs.get('headers') or {})
        kwargs['headers'] = headers
        return super()._download_webpage_handle(
            *args, **kwargs)

    @staticmethod
    def _get_mso_headers(mso_info):
        # Not needed currently
        return {}

    @staticmethod
    def _get_mvpd_resource(provider_id, title, guid, rating):
        channel = etree.Element('channel')
        channel_title = etree.SubElement(channel, 'title')
        channel_title.text = provider_id
        item = etree.SubElement(channel, 'item')
        resource_title = etree.SubElement(item, 'title')
        resource_title.text = title
        resource_guid = etree.SubElement(item, 'guid')
        resource_guid.text = guid
        resource_rating = etree.SubElement(item, 'media:rating')
        resource_rating.attrib = {'scheme': 'urn:v-chip'}
        resource_rating.text = rating
        return '<rss version="2.0" xmlns:media="http://search.yahoo.com/mrss/">' + etree.tostring(channel).decode() + '</rss>'

    def _extract_mvpd_auth(self, url, video_id, requestor_id, resource, software_statement):
        mso_id = self.get_param('ap_mso')
        if mso_id:
            mso_info = MSO_INFO[mso_id]
        else:
            mso_info = {}

        def xml_text(xml_str, tag):
            return self._search_regex(
                f'<{tag}>(.+?)</{tag}>', xml_str, tag)

        def is_expired(token, date_ele):
            token_expires = unified_timestamp(re.sub(r'[_ ]GMT', '', xml_text(token, date_ele)))
            return token_expires and token_expires <= int(time.time())

        def post_form(form_page_res, note, data={}, validate_url=False):
            form_page, urlh = form_page_res
            post_url = self._html_search_regex(r'<form[^>]+action=(["\'])(?P<url>.+?)\1', form_page, 'post url', group='url')
            if not re.match(r'https?://', post_url):
                post_url = urllib.parse.urljoin(urlh.url, post_url)
            if validate_url:
                # This request is submitting credentials so we should validate it when possible
                url_parsed = urllib.parse.urlparse(post_url)
                expected_hostname = mso_info.get('login_hostname')
                if expected_hostname and expected_hostname != url_parsed.hostname:
                    raise ExtractorError(
                        f'Unexpected login URL hostname; expected "{expected_hostname}" but got '
                        f'"{url_parsed.hostname}". Aborting before submitting credentials')
                if url_parsed.scheme != 'https':
                    self.write_debug('Upgrading login URL scheme to https')
                    post_url = urllib.parse.urlunparse(url_parsed._replace(scheme='https'))
            form_data = self._hidden_inputs(form_page)
            form_data.update(data)
            return self._download_webpage_handle(
                post_url, video_id, note, data=urlencode_postdata(form_data), headers={
                    **self._get_mso_headers(mso_info),
                    'Content-Type': 'application/x-www-form-urlencoded',
                })

        def raise_mvpd_required():
            raise ExtractorError(
                'This video is only available for users of participating TV providers. '
                'Use --ap-mso to specify Adobe Pass Multiple-system operator Identifier '
                'and --ap-username and --ap-password or --netrc to provide account credentials.', expected=True)

        def extract_redirect_url(html, url=None, fatal=False):
            # TODO: eliminate code duplication with generic extractor and move
            # redirection code into _download_webpage_handle
            REDIRECT_REGEX = r'[0-9]{,2};\s*(?:URL|url)=\'?([^\'"]+)'
            redirect_url = self._search_regex(
                r'(?i)<meta\s+(?=(?:[a-z-]+="[^"]+"\s+)*http-equiv="refresh")'
                rf'(?:[a-z-]+="[^"]+"\s+)*?content="{REDIRECT_REGEX}',
                html, 'meta refresh redirect',
                default=NO_DEFAULT if fatal else None, fatal=fatal)
            if not redirect_url:
                return None
            if url:
                redirect_url = urllib.parse.urljoin(url, unescapeHTML(redirect_url))
            return redirect_url

        mvpd_headers = {
            'ap_42': 'anonymous',
            'ap_11': 'Linux i686',
            'ap_z': self._USER_AGENT,
            'User-Agent': self._USER_AGENT,
        }

        guid = xml_text(resource, 'guid') if '<' in resource else resource
        for _ in range(2):
            requestor_info = self.cache.load(self._MVPD_CACHE, requestor_id) or {}
            authn_token = requestor_info.get('authn_token')
            if authn_token and is_expired(authn_token, 'simpleTokenExpires'):
                authn_token = None
            if not authn_token:
                if not mso_id:
                    raise_mvpd_required()
                username, password = self._get_login_info('ap_username', 'ap_password', mso_id)
                if not username or not password:
                    raise_mvpd_required()

                device_info, urlh = self._download_json_handle(
                    'https://sp.auth.adobe.com/indiv/devices',
                    video_id, 'Registering device with Adobe',
                    data=json.dumps({'fingerprint': uuid.uuid4().hex}).encode(),
                    headers={'Content-Type': 'application/json; charset=UTF-8'})

                device_id = device_info['deviceId']
                mvpd_headers['pass_sfp'] = urlh.get_header('pass_sfp')
                mvpd_headers['Ap_21'] = device_id

                registration = self._download_json(
                    'https://sp.auth.adobe.com/o/client/register',
                    video_id, 'Registering client with Adobe',
                    data=json.dumps({'software_statement': software_statement}).encode(),
                    headers={'Content-Type': 'application/json; charset=UTF-8'})

                access_token = self._download_json(
                    'https://sp.auth.adobe.com/o/client/token', video_id,
                    'Obtaining access token', data=urlencode_postdata({
                        'grant_type': 'client_credentials',
                        'client_id': registration['client_id'],
                        'client_secret': registration['client_secret'],
                    }),
                    headers={
                        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
                    })['access_token']
                mvpd_headers['Authorization'] = f'Bearer {access_token}'

                reg_code = self._download_json(
                    f'https://sp.auth.adobe.com/reggie/v1/{requestor_id}/regcode',
                    video_id, 'Obtaining registration code',
                    data=urlencode_postdata({
                        'requestor': requestor_id,
                        'deviceId': device_id,
                        'format': 'json',
                    }),
                    headers={
                        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
                        'Authorization': f'Bearer {access_token}',
                    })['code']

                provider_redirect_page_res = self._download_webpage_handle(
                    self._SERVICE_PROVIDER_TEMPLATE % 'authenticate/saml', video_id,
                    'Downloading Provider Redirect Page', query={
                        'noflash': 'true',
                        'mso_id': mso_id,
                        'requestor_id': requestor_id,
                        'no_iframe': 'false',
                        'domain_name': 'adobe.com',
                        'redirect_url': url,
                        'reg_code': reg_code,
                    }, headers=self._get_mso_headers(mso_info))

                if mso_id == 'Comcast_SSO':
                    # Comcast page flow varies by video site and whether you
                    # are on Comcast's network.
                    provider_redirect_page, urlh = provider_redirect_page_res
                    if 'automatically signing you in' in provider_redirect_page:
                        oauth_redirect_url = self._html_search_regex(
                            r'window\.location\s*=\s*[\'"]([^\'"]+)',
                            provider_redirect_page, 'oauth redirect')
                        self._download_webpage(
                            oauth_redirect_url, video_id, 'Confirming auto login')
                    elif 'automatically signed in with' in provider_redirect_page:
                        # Seems like comcast is rolling up new way of automatically signing customers
                        oauth_redirect_url = self._html_search_regex(
                            r'continue:\s*"(https://oauth\.xfinity\.com/oauth/authorize\?.+)"', provider_redirect_page,
                            'oauth redirect (signed)')
                        # Just need to process the request. No useful data comes back
                        self._download_webpage(oauth_redirect_url, video_id, 'Confirming auto login')
                    else:
                        if '<form name="signin"' in provider_redirect_page:
                            provider_login_page_res = provider_redirect_page_res
                        elif 'http-equiv="refresh"' in provider_redirect_page:
                            oauth_redirect_url = extract_redirect_url(
                                provider_redirect_page, fatal=True)
                            provider_login_page_res = self._download_webpage_handle(
                                oauth_redirect_url, video_id, self._DOWNLOADING_LOGIN_PAGE,
                                headers=self._get_mso_headers(mso_info))
                        else:
                            provider_login_page_res = post_form(
                                provider_redirect_page_res,
                                self._DOWNLOADING_LOGIN_PAGE)

                        mvpd_confirm_page_res = post_form(
                            provider_login_page_res, 'Logging in', {
                                mso_info['username_field']: username,
                                mso_info['password_field']: password,
                            }, validate_url=True)
                        mvpd_confirm_page, urlh = mvpd_confirm_page_res
                        if '<button class="submit" value="Resume">Resume</button>' in mvpd_confirm_page:
                            post_form(mvpd_confirm_page_res, 'Confirming Login')
                elif mso_id == 'Philo':
                    # Philo has very unique authentication method
                    self._request_webpage(
                        'https://idp.philo.com/auth/init/login_code', video_id,
                        'Requesting Philo auth code', data=json.dumps({
                            'ident': username,
                            'device': 'web',
                            'send_confirm_link': False,
                            'send_token': True,
                            'device_ident': f'web-{uuid.uuid4().hex}',
                            'include_login_link': True,
                        }).encode(), headers={
                            'Content-Type': 'application/json',
                            'Accept': 'application/json',
                        })

                    philo_code = getpass.getpass('Type auth code you have received [Return]: ')
                    self._request_webpage(
                        'https://idp.philo.com/auth/update/login_code', video_id,
                        'Submitting token', data=json.dumps({'token': philo_code}).encode(),
                        headers={
                            'Content-Type': 'application/json',
                            'Accept': 'application/json',
                        })

                    mvpd_confirm_page_res = self._download_webpage_handle('https://idp.philo.com/idp/submit', video_id, 'Confirming Philo Login')
                    post_form(mvpd_confirm_page_res, 'Confirming Login')
                elif mso_id == 'Verizon':
                    # In general, if you're connecting from a Verizon-assigned IP,
                    # you will not actually pass your credentials.
                    provider_redirect_page, urlh = provider_redirect_page_res
                    # From non-Verizon IP, still gave 'Please wait', but noticed N==Y; will need to try on Verizon IP
                    if 'Please wait ...' in provider_redirect_page and '\'N\'== "Y"' not in provider_redirect_page:
                        saml_redirect_url = self._html_search_regex(
                            r'self\.parent\.location=(["\'])(?P<url>.+?)\1',
                            provider_redirect_page,
                            'SAML Redirect URL', group='url')
                        saml_login_page = self._download_webpage(
                            saml_redirect_url, video_id,
                            'Downloading SAML Login Page')
                    elif 'Verizon FiOS - sign in' in provider_redirect_page:
                        # FXNetworks from non-Verizon IP
                        saml_login_page_res = post_form(
                            provider_redirect_page_res, 'Logging in', {
                                mso_info['username_field']: username,
                                mso_info['password_field']: password,
                            }, validate_url=True)
                        saml_login_page, urlh = saml_login_page_res
                        if 'Please try again.' in saml_login_page:
                            raise ExtractorError(
                                'We\'re sorry, but either the User ID or Password entered is not correct.')
                    else:
                        # ABC from non-Verizon IP
                        saml_redirect_url = self._html_search_regex(
                            r'var\surl\s*=\s*(["\'])(?P<url>.+?)\1',
                            provider_redirect_page,
                            'SAML Redirect URL', group='url')
                        saml_redirect_url = saml_redirect_url.replace(r'\/', '/')
                        saml_redirect_url = saml_redirect_url.replace(r'\-', '-')
                        saml_redirect_url = saml_redirect_url.replace(r'\x26', '&')
                        saml_login_page = self._download_webpage(
                            saml_redirect_url, video_id,
                            'Downloading SAML Login Page')
                        saml_login_page, urlh = post_form(
                            [saml_login_page, saml_redirect_url], 'Logging in', {
                                mso_info['username_field']: username,
                                mso_info['password_field']: password,
                            }, validate_url=True)
                        if 'Please try again.' in saml_login_page:
                            raise ExtractorError(
                                'Failed to login, incorrect User ID or Password.')
                    saml_login_url = self._search_regex(
                        r'xmlHttp\.open\("POST"\s*,\s*(["\'])(?P<url>.+?)\1',
                        saml_login_page, 'SAML Login URL', group='url')
                    saml_response_json = self._download_json(
                        saml_login_url, video_id, 'Downloading SAML Response',
                        headers={'Content-Type': 'text/xml'})
                    self._download_webpage(
                        saml_response_json['targetValue'], video_id,
                        'Confirming Login', data=urlencode_postdata({
                            'SAMLResponse': saml_response_json['SAMLResponse'],
                            'RelayState': saml_response_json['RelayState'],
                        }), headers={
                            'Content-Type': 'application/x-www-form-urlencoded',
                        })
                elif mso_id in ('Spectrum', 'Charter_Direct'):
                    # Spectrum's login for is dynamically loaded via JS so we need to hardcode the flow
                    # as a one-off implementation.
                    provider_redirect_page, urlh = provider_redirect_page_res
                    provider_login_page_res = post_form(
                        provider_redirect_page_res, self._DOWNLOADING_LOGIN_PAGE)
                    saml_login_page, urlh = provider_login_page_res
                    relay_state = self._search_regex(
                        r'RelayState\s*=\s*"(?P<relay>.+?)";',
                        saml_login_page, 'RelayState', group='relay')
                    saml_request = self._search_regex(
                        r'SAMLRequest\s*=\s*"(?P<saml_request>.+?)";',
                        saml_login_page, 'SAMLRequest', group='saml_request')
                    login_json = {
                        mso_info['username_field']: username,
                        mso_info['password_field']: password,
                        'RelayState': relay_state,
                        'SAMLRequest': saml_request,
                    }
                    saml_response_json = self._download_json(
                        'https://tveauthn.spectrum.net/tveauthentication/api/v1/manualAuth', video_id,
                        'Downloading SAML Response',
                        data=json.dumps(login_json).encode(),
                        headers={
                            'Content-Type': 'application/json',
                            'Accept': 'application/json',
                        })
                    self._download_webpage(
                        saml_response_json['SAMLRedirectUri'], video_id,
                        'Confirming Login', data=urlencode_postdata({
                            'SAMLResponse': saml_response_json['SAMLResponse'],
                            'RelayState': relay_state,
                        }), headers={
                            'Content-Type': 'application/x-www-form-urlencoded',
                        })
                elif mso_id == 'slingtv':
                    # SlingTV has a meta-refresh based authentication, but also
                    # looks at the tab history to count the number of times the
                    # browser has been on a page

                    first_bookend_page, urlh = provider_redirect_page_res

                    hidden_data = self._hidden_inputs(first_bookend_page)
                    hidden_data['history'] = 1

                    provider_login_page_res = self._download_webpage_handle(
                        urlh.url, video_id, 'Sending first bookend',
                        query=hidden_data)

                    provider_association_redirect, urlh = post_form(
                        provider_login_page_res, 'Logging in', {
                            mso_info['username_field']: username,
                            mso_info['password_field']: password,
                        }, validate_url=True)

                    provider_refresh_redirect_url = extract_redirect_url(
                        provider_association_redirect, url=urlh.url)

                    last_bookend_page, urlh = self._download_webpage_handle(
                        provider_refresh_redirect_url, video_id,
                        'Downloading Auth Association Redirect Page')
                    hidden_data = self._hidden_inputs(last_bookend_page)
                    hidden_data['history'] = 3

                    mvpd_confirm_page_res = self._download_webpage_handle(
                        urlh.url, video_id, 'Sending final bookend',
                        query=hidden_data)

                    post_form(mvpd_confirm_page_res, 'Confirming Login')
                elif mso_id == 'Suddenlink':
                    # Suddenlink is similar to SlingTV in using a tab history count and a meta refresh,
                    # but they also do a dynmaic redirect using javascript that has to be followed as well
                    first_bookend_page, urlh = post_form(
                        provider_redirect_page_res, 'Pressing Continue...')

                    hidden_data = self._hidden_inputs(first_bookend_page)
                    hidden_data['history_val'] = 1

                    provider_login_redirect_page_res = self._download_webpage_handle(
                        urlh.url, video_id, 'Sending First Bookend',
                        query=hidden_data)

                    provider_login_redirect_page, urlh = provider_login_redirect_page_res

                    # Some website partners seem to not have the extra ajaxurl redirect step, so we check if we already
                    # have the login prompt or not
                    if 'id="password" type="password" name="password"' in provider_login_redirect_page:
                        provider_login_page_res = provider_login_redirect_page_res
                    else:
                        provider_tryauth_url = self._html_search_regex(
                            r'url:\s*[\'"]([^\'"]+)', provider_login_redirect_page, 'ajaxurl')
                        provider_tryauth_page = self._download_webpage(
                            provider_tryauth_url, video_id, 'Submitting TryAuth',
                            query=hidden_data)

                        provider_login_page_res = self._download_webpage_handle(
                            f'https://authorize.suddenlink.net/saml/module.php/authSynacor/login.php?AuthState={provider_tryauth_page}',
                            video_id, 'Getting Login Page',
                            query=hidden_data)

                    provider_association_redirect, urlh = post_form(
                        provider_login_page_res, 'Logging in', {
                            mso_info['username_field']: username,
                            mso_info['password_field']: password,
                        }, validate_url=True)

                    provider_refresh_redirect_url = extract_redirect_url(
                        provider_association_redirect, url=urlh.url)

                    last_bookend_page, urlh = self._download_webpage_handle(
                        provider_refresh_redirect_url, video_id,
                        'Downloading Auth Association Redirect Page')

                    hidden_data = self._hidden_inputs(last_bookend_page)
                    hidden_data['history_val'] = 3

                    mvpd_confirm_page_res = self._download_webpage_handle(
                        urlh.url, video_id, 'Sending Final Bookend',
                        query=hidden_data)

                    post_form(mvpd_confirm_page_res, 'Confirming Login')
                elif mso_id == 'Fubo':
                    _, urlh = provider_redirect_page_res

                    fubo_response = self._download_json(
                        'https://api.fubo.tv/partners/tve/connect', video_id,
                        'Authenticating with Fubo', 'Unable to authenticate with Fubo',
                        query=parse_qs(urlh.url), data=json.dumps({
                            'username': username,
                            'password': password,
                        }).encode(), headers={
                            'Accept': 'application/json',
                            'Content-Type': 'application/json',
                        })

                    self._request_webpage(
                        'https://sp.auth.adobe.com/adobe-services/oauth2', video_id,
                        'Authenticating with Adobe', 'Failed to authenticate with Adobe',
                        query={
                            'code': fubo_response['code'],
                            'state': fubo_response['state'],
                        })
                else:
                    # Some providers (e.g. DIRECTV NOW) have another meta refresh
                    # based redirect that should be followed.
                    provider_redirect_page, urlh = provider_redirect_page_res
                    provider_refresh_redirect_url = extract_redirect_url(
                        provider_redirect_page, url=urlh.url)
                    if provider_refresh_redirect_url:
                        provider_redirect_page_res = self._download_webpage_handle(
                            provider_refresh_redirect_url, video_id,
                            'Downloading Provider Redirect Page (meta refresh)')
                    provider_login_page_res = post_form(
                        provider_redirect_page_res, self._DOWNLOADING_LOGIN_PAGE)
                    form_data = {
                        mso_info.get('username_field', 'username'): username,
                        mso_info.get('password_field', 'password'): password,
                    }
                    if mso_id in ('Cablevision', 'AlticeOne'):
                        form_data['_eventId_proceed'] = ''
                    mvpd_confirm_page_res = post_form(
                        provider_login_page_res, 'Logging in', form_data, validate_url=True)
                    if mso_id != 'Rogers':
                        post_form(mvpd_confirm_page_res, 'Confirming Login')

                try:
                    session = self._download_webpage(
                        self._SERVICE_PROVIDER_TEMPLATE % 'session', video_id,
                        'Retrieving Session', data=urlencode_postdata({
                            '_method': 'GET',
                            'requestor_id': requestor_id,
                            'reg_code': reg_code,
                        }), headers=mvpd_headers)
                except ExtractorError as e:
                    if not mso_id and isinstance(e.cause, HTTPError) and e.cause.status == 401:
                        raise_mvpd_required()
                    raise
                if '<pendingLogout' in session:
                    self.cache.store(self._MVPD_CACHE, requestor_id, {})
                    continue
                authn_token = unescapeHTML(xml_text(session, 'authnToken'))
                requestor_info['authn_token'] = authn_token
                self.cache.store(self._MVPD_CACHE, requestor_id, requestor_info)

            authz_token = requestor_info.get(guid)
            if authz_token and is_expired(authz_token, 'simpleTokenTTL'):
                authz_token = None
            if not authz_token:
                authorize = self._download_webpage(
                    self._SERVICE_PROVIDER_TEMPLATE % 'authorize', video_id,
                    'Retrieving Authorization Token', data=urlencode_postdata({
                        'resource_id': resource,
                        'requestor_id': requestor_id,
                        'authentication_token': authn_token,
                        'mso_id': xml_text(authn_token, 'simpleTokenMsoID'),
                        'userMeta': '1',
                    }), headers=mvpd_headers)
                if '<pendingLogout' in authorize:
                    self.cache.store(self._MVPD_CACHE, requestor_id, {})
                    continue
                if '<error' in authorize:
                    raise ExtractorError(xml_text(authorize, 'details'), expected=True)
                authz_token = unescapeHTML(xml_text(authorize, 'authzToken'))
                requestor_info[guid] = authz_token
                self.cache.store(self._MVPD_CACHE, requestor_id, requestor_info)

            mvpd_headers.update({
                'ap_19': xml_text(authn_token, 'simpleSamlNameID'),
                'ap_23': xml_text(authn_token, 'simpleSamlSessionIndex'),
            })

            short_authorize = self._download_webpage(
                self._SERVICE_PROVIDER_TEMPLATE % 'shortAuthorize',
                video_id, 'Retrieving Media Token', data=urlencode_postdata({
                    'authz_token': authz_token,
                    'requestor_id': requestor_id,
                    'session_guid': xml_text(authn_token, 'simpleTokenAuthenticationGuid'),
                    'hashed_guid': 'false',
                }), headers=mvpd_headers)
            if '<pendingLogout' in short_authorize:
                self.cache.store(self._MVPD_CACHE, requestor_id, {})
                continue
            return short_authorize



from .common import InfoExtractor
from ..utils import (
    int_or_none,
    traverse_obj,
    try_get,
    urljoin,
)


class MxplayerIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?mxplayer\.in/(?P<type>movie|show/[-\w]+/[-\w]+)/(?P<display_id>[-\w]+)-(?P<id>\w+)'
    _TESTS = [{
        'url': 'https://www.mxplayer.in/show/watch-my-girlfriend-is-an-alien-hindi-dubbed/season-1/episode-1-online-9d2013d31d5835bb8400e3b3c5e7bb72',
        'info_dict': {
            'id': '9d2013d31d5835bb8400e3b3c5e7bb72',
            'display_id': 'episode-1-online',
            'ext': 'mp4',
            'title': 'Episode 1',
            'description': 'md5:62ed43eb9fec5efde5cf3bd1040b7670',
            'season_number': 1,
            'episode_number': 1,
            'duration': 2451,
            'season': 'Season 1',
            'series': 'My Girlfriend Is An Alien (Hindi Dubbed)',
            'episode': 'Episode 1',
        },
        'params': {
            'format': 'bv',
            'skip_download': True,
        },
    }, {
        'url': 'https://www.mxplayer.in/movie/watch-knock-knock-hindi-dubbed-movie-online-b9fa28df3bfb8758874735bbd7d2655a?watch=true',
        'info_dict': {
            'id': 'b9fa28df3bfb8758874735bbd7d2655a',
            'display_id': 'episode-1-online',
            'ext': 'mp4',
            'title': 'Knock Knock (Hindi Dubbed)',
            'description': 'md5:4160f2dfc3b87c524261366f6b736329',
            'duration': 5970,
        },
        'params': {
            'format': 'bv',
            'skip_download': True,
        },
        'skip': 'No longer available',
    }, {
        'url': 'https://www.mxplayer.in/show/watch-shaitaan/season-1/the-infamous-taxi-gang-of-meerut-online-45055d5bcff169ad48f2ad7552a83d6c',
        'info_dict': {
            'id': '45055d5bcff169ad48f2ad7552a83d6c',
            'ext': 'mp4',
            'title': 'The infamous taxi gang of Meerut',
            'description': 'md5:033a0a7e3fd147be4fb7e07a01a3dc28',
            'season_number': 1,
            'episode_number': 1,
            'duration': 2332,
            'season': 'Season 1',
            'series': 'Shaitaan',
            'episode': 'Episode 1',
        },
        'params': {
            'format': 'best',
            'skip_download': True,
        },
        'skip': 'No longer available.',
    }, {
        'url': 'https://www.mxplayer.in/show/watch-aashram/chapter-1/duh-swapna-online-d445579792b0135598ba1bc9088a84cb',
        'info_dict': {
            'id': 'd445579792b0135598ba1bc9088a84cb',
            'display_id': 'duh-swapna-online',
            'ext': 'mp4',
            'title': 'Duh Swapna',
            'description': 'md5:35ff39c4bdac403c53be1e16a04192d8',
            'season_number': 1,
            'episode_number': 3,
            'duration': 2568,
            'season': 'Season 1',
            'series': 'Aashram',
            'episode': 'Episode 3',
        },
        'params': {
            'format': 'bv',
            'skip_download': True,
        },
    }, {
        'url': 'https://www.mxplayer.in/show/watch-dangerous/season-1/chapter-1-online-5a351b4f9fb69436f6bd6ae3a1a75292',
        'info_dict': {
            'id': '5a351b4f9fb69436f6bd6ae3a1a75292',
            'display_id': 'chapter-1-online',
            'ext': 'mp4',
            'title': 'Chapter 1',
            'description': 'md5:233886b8598bc91648ac098abe1d288f',
            'season_number': 1,
            'episode_number': 1,
            'duration': 1305,
            'season': 'Season 1',
            'series': 'Dangerous',
            'episode': 'Episode 1',
        },
        'params': {
            'format': 'bv',
            'skip_download': True,
        },
    }, {
        'url': 'https://www.mxplayer.in/movie/watch-the-attacks-of-2611-movie-online-0452f0d80226c398d63ce7e3ea40fa2d',
        'info_dict': {
            'id': '0452f0d80226c398d63ce7e3ea40fa2d',
            'ext': 'mp4',
            'title': 'The Attacks of 26/11',
            'description': 'md5:689bacd29e97b3f31eaf519eb14127e5',
            'duration': 6085,
        },
        'params': {
            'format': 'best',
            'skip_download': True,
        },
        'skip': 'No longer available. Cannot be played on browser',
    }, {
        'url': 'https://www.mxplayer.in/movie/watch-kitne-door-kitne-paas-movie-online-a9e9c76c566205955f70d8b2cb88a6a2',
        'info_dict': {
            'id': 'a9e9c76c566205955f70d8b2cb88a6a2',
            'display_id': 'watch-kitne-door-kitne-paas-movie-online',
            'title': 'Kitne Door Kitne Paas',
            'duration': 8458,
            'ext': 'mp4',
            'description': 'md5:fb825f3c542513088024dcafef0921b4',
        },
        'params': {
            'format': 'bv',
            'skip_download': True,
        },
    }, {
        'url': 'https://www.mxplayer.in/show/watch-ek-thi-begum-hindi/season-2/game-of-power-online-5e5305c28f1409847cdc4520b6ad77cf',
        'info_dict': {
            'id': '5e5305c28f1409847cdc4520b6ad77cf',
            'display_id': 'game-of-power-online',
            'title': 'Game Of Power',
            'duration': 1845,
            'ext': 'mp4',
            'description': 'md5:1d0948d2a5312d7013792d53542407f9',
            'series': 'Ek Thi Begum (Hindi)',
            'season': 'Season 2',
            'season_number': 2,
            'episode': 'Episode 2',
            'episode_number': 2,
        },
        'params': {
            'format': 'bv',
            'skip_download': True,
        },
    }, {
        'url': 'https://www.mxplayer.in/movie/watch-deewane-huye-paagal-movie-online-4f9175c40a11c3994182a65afdd37ec6?watch=true',
        'info_dict': {
            'id': '4f9175c40a11c3994182a65afdd37ec6',
            'display_id': 'watch-deewane-huye-paagal-movie-online',
            'title': 'Deewane Huye Paagal',
            'duration': 9037,
            'ext': 'mp4',
            'description': 'md5:d17bd5c651016c4ed2e6f8a4ace15534',
        },
        'params': {'skip_download': 'm3u8'},
    }]

    def _real_extract(self, url):
        video_type, display_id, video_id = self._match_valid_url(url).group('type', 'display_id', 'id')
        if 'show' in video_type:
            video_type = 'episode'

        data_json = self._download_json(
            f'https://api.mxplay.com/v1/web/detail/video?type={video_type}&id={video_id}', display_id)

        formats, subtitles = [], {}
        m3u8_url = urljoin('https://llvod.mxplay.com/', traverse_obj(
            data_json, ('stream', (('thirdParty', 'hlsUrl'), ('hls', 'high'))), get_all=False))
        if m3u8_url:
            formats, subtitles = self._extract_m3u8_formats_and_subtitles(m3u8_url, display_id, 'mp4', fatal=False)
        mpd_url = urljoin('https://llvod.mxplay.com/', traverse_obj(
            data_json, ('stream', (('thirdParty', 'dashUrl'), ('dash', 'high'))), get_all=False))
        if mpd_url:
            fmts, subs = self._extract_mpd_formats_and_subtitles(mpd_url, display_id, fatal=False)
            formats.extend(fmts)
            self._merge_subtitles(subs, target=subtitles)

        season = traverse_obj(data_json, ('container', 'title'))
        return {
            'id': video_id,
            'title': data_json.get('title'),
            'formats': formats,
            'subtitles': subtitles,
            'display_id': display_id,
            'duration': data_json.get('duration'),
            'series': traverse_obj(data_json, ('container', 'container', 'title')),
            'description': data_json.get('description'),
            'season': season,
            'season_number': int_or_none(
                self._search_regex(r'Season (\d+)', season, 'Season Number', default=None)),
            'episode_number': data_json.get('sequence') or None,
        }


class MxplayerShowIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?mxplayer\.in/show/(?P<display_id>[-\w]+)-(?P<id>\w+)/?(?:$|[#?])'
    _TESTS = [{
        'url': 'https://www.mxplayer.in/show/watch-chakravartin-ashoka-samrat-series-online-a8f44e3cc0814b5601d17772cedf5417',
        'playlist_mincount': 440,
        'info_dict': {
            'id': 'a8f44e3cc0814b5601d17772cedf5417',
            'title': 'Watch Chakravartin Ashoka Samrat Series Online',
        },
    }]

    _API_SHOW_URL = 'https://api.mxplay.com/v1/web/detail/tab/tvshowseasons?type=tv_show&id={}&device-density=2&platform=com.mxplay.desktop&content-languages=hi,en'
    _API_EPISODES_URL = 'https://api.mxplay.com/v1/web/detail/tab/tvshowepisodes?type=season&id={}&device-density=1&platform=com.mxplay.desktop&content-languages=hi,en&{}'

    def _entries(self, show_id):
        show_json = self._download_json(
            self._API_SHOW_URL.format(show_id),
            video_id=show_id, headers={'Referer': 'https://mxplayer.in'})
        page_num = 0
        for season in show_json.get('items') or []:
            season_id = try_get(season, lambda x: x['id'], str)
            next_url = ''
            while next_url is not None:
                page_num += 1
                season_json = self._download_json(
                    self._API_EPISODES_URL.format(season_id, next_url),
                    video_id=season_id,
                    headers={'Referer': 'https://mxplayer.in'},
                    note=f'Downloading JSON metadata page {page_num}')
                for episode in season_json.get('items') or []:
                    video_url = episode['webUrl']
                    yield self.url_result(
                        f'https://mxplayer.in{video_url}',
                        ie=MxplayerIE.ie_key(), video_id=video_url.split('-')[-1])
                next_url = season_json.get('next')

    def _real_extract(self, url):
        display_id, show_id = self._match_valid_url(url).groups()
        return self.playlist_result(
            self._entries(show_id), playlist_id=show_id,
            playlist_title=display_id.replace('-', ' ').title())



import re

from .common import InfoExtractor
from ..utils import (
    int_or_none,
    join_nonempty,
    parse_duration,
    urljoin,
    xpath_element,
    xpath_text,
)


class HBOBaseIE(InfoExtractor):
    _FORMATS_INFO = {
        'pro7': {
            'width': 1280,
            'height': 720,
        },
        '1920': {
            'width': 1280,
            'height': 720,
        },
        'pro6': {
            'width': 768,
            'height': 432,
        },
        '640': {
            'width': 768,
            'height': 432,
        },
        'pro5': {
            'width': 640,
            'height': 360,
        },
        'highwifi': {
            'width': 640,
            'height': 360,
        },
        'high3g': {
            'width': 640,
            'height': 360,
        },
        'medwifi': {
            'width': 400,
            'height': 224,
        },
        'med3g': {
            'width': 400,
            'height': 224,
        },
    }

    def _extract_info(self, url, display_id):
        video_data = self._download_xml(url, display_id)
        video_id = xpath_text(video_data, 'id', fatal=True)
        episode_title = title = xpath_text(video_data, 'title', fatal=True)
        series = xpath_text(video_data, 'program')
        if series:
            title = f'{series} - {title}'

        formats = []
        for source in xpath_element(video_data, 'videos', 'sources', True):
            if source.tag == 'size':
                path = xpath_text(source, './/path')
                if not path:
                    continue
                width = source.attrib.get('width')
                format_info = self._FORMATS_INFO.get(width, {})
                height = format_info.get('height')
                fmt = {
                    'url': path,
                    'format_id': join_nonempty('http'. height and f'{height}p'),
                    'width': format_info.get('width'),
                    'height': height,
                }
                rtmp = re.search(r'^(?P<url>rtmpe?://[^/]+/(?P<app>.+))/(?P<playpath>mp4:.+)$', path)
                if rtmp:
                    fmt.update({
                        'url': rtmp.group('url'),
                        'play_path': rtmp.group('playpath'),
                        'app': rtmp.group('app'),
                        'ext': 'flv',
                        'format_id': fmt['format_id'].replace('http', 'rtmp'),
                    })
                formats.append(fmt)
            else:
                video_url = source.text
                if not video_url:
                    continue
                if source.tag == 'tarball':
                    formats.extend(self._extract_m3u8_formats(
                        video_url.replace('.tar', '/base_index_w8.m3u8'),
                        video_id, 'mp4', 'm3u8_native', m3u8_id='hls', fatal=False))
                elif source.tag == 'hls':
                    m3u8_formats = self._extract_m3u8_formats(
                        video_url.replace('.tar', '/base_index.m3u8'),
                        video_id, 'mp4', 'm3u8_native', m3u8_id='hls', fatal=False)
                    for f in m3u8_formats:
                        if f.get('vcodec') == 'none' and not f.get('tbr'):
                            f['tbr'] = int_or_none(self._search_regex(
                                r'-(\d+)k/', f['url'], 'tbr', default=None))
                    formats.extend(m3u8_formats)
                elif source.tag == 'dash':
                    formats.extend(self._extract_mpd_formats(
                        video_url.replace('.tar', '/manifest.mpd'),
                        video_id, mpd_id='dash', fatal=False))
                else:
                    format_info = self._FORMATS_INFO.get(source.tag, {})
                    formats.append({
                        'format_id': f'http-{source.tag}',
                        'url': video_url,
                        'width': format_info.get('width'),
                        'height': format_info.get('height'),
                    })

        thumbnails = []
        card_sizes = xpath_element(video_data, 'titleCardSizes')
        if card_sizes is not None:
            for size in card_sizes:
                path = xpath_text(size, 'path')
                if not path:
                    continue
                width = int_or_none(size.get('width'))
                thumbnails.append({
                    'id': width,
                    'url': path,
                    'width': width,
                })

        subtitles = None
        caption_url = xpath_text(video_data, 'captionUrl')
        if caption_url:
            subtitles = {
                'en': [{
                    'url': caption_url,
                    'ext': 'ttml',
                }],
            }

        return {
            'id': video_id,
            'title': title,
            'duration': parse_duration(xpath_text(video_data, 'duration/tv14')),
            'series': series,
            'episode': episode_title,
            'formats': formats,
            'thumbnails': thumbnails,
            'subtitles': subtitles,
        }


class HBOIE(HBOBaseIE):
    IE_NAME = 'hbo'
    _VALID_URL = r'https?://(?:www\.)?hbo\.com/(?:video|embed)(?:/[^/]+)*/(?P<id>[^/?#]+)'
    _TEST = {
        'url': 'https://www.hbo.com/video/game-of-thrones/seasons/season-8/videos/trailer',
        'md5': '8126210656f433c452a21367f9ad85b3',
        'info_dict': {
            'id': '22113301',
            'ext': 'mp4',
            'title': 'Game of Thrones - Trailer',
        },
        'expected_warnings': ['Unknown MIME type application/mp4 in DASH manifest'],
    }

    def _real_extract(self, url):
        display_id = self._match_id(url)
        webpage = self._download_webpage(url, display_id)
        location_path = self._parse_json(self._html_search_regex(
            r'data-state="({.+?})"', webpage, 'state'), display_id)['video']['locationUrl']
        return self._extract_info(urljoin(url, location_path), display_id)



import hashlib
import hmac
import itertools
import json
import urllib.parse

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    int_or_none,
    join_nonempty,
    try_call,
    url_or_none,
    urljoin,
)
from ..utils.traversal import traverse_obj


class ZingMp3BaseIE(InfoExtractor):
    _VALID_URL_TMPL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<type>(?:%s))/[^/?#]+/(?P<id>\w+)(?:\.html|\?)'
    _GEO_COUNTRIES = ['VN']
    _DOMAIN = 'https://zingmp3.vn'
    _PER_PAGE = 50
    _API_SLUGS = {
        # Audio/video
        'bai-hat': '/api/v2/page/get/song',
        'embed': '/api/v2/page/get/song',
        'video-clip': '/api/v2/page/get/video',
        'lyric': '/api/v2/lyric/get/lyric',
        'song-streaming': '/api/v2/song/get/streaming',
        'liveradio': '/api/v2/livestream/get/info',
        'eps': '/api/v2/page/get/podcast-episode',
        'episode-streaming': '/api/v2/podcast/episode/get/streaming',
        # Playlist
        'playlist': '/api/v2/page/get/playlist',
        'album': '/api/v2/page/get/playlist',
        'pgr': '/api/v2/page/get/podcast-program',
        'pgr-list': '/api/v2/podcast/episode/get/list',
        'cgr': '/api/v2/page/get/podcast-category',
        'cgr-list': '/api/v2/podcast/program/get/list-by-cate',
        'cgrs': '/api/v2/page/get/podcast-categories',
        # Chart
        'zing-chart': '/api/v2/page/get/chart-home',
        'zing-chart-tuan': '/api/v2/page/get/week-chart',
        'moi-phat-hanh': '/api/v2/page/get/newrelease-chart',
        'the-loai-video': '/api/v2/video/get/list',
        # User
        'info-artist': '/api/v2/page/get/artist',
        'user-list-song': '/api/v2/song/get/list',
        'user-list-video': '/api/v2/video/get/list',
        'hub': '/api/v2/page/get/hub-detail',
        'new-release': '/api/v2/chart/get/new-release',
        'top100': '/api/v2/page/get/top-100',
        'podcast-new': '/api/v2/podcast/program/get/list-by-type',
        'top-podcast': '/api/v2/podcast/program/get/top-episode',
    }

    def _api_url(self, url_type, params):
        api_slug = self._API_SLUGS[url_type]
        params.update({'ctime': '1'})
        sha256 = hashlib.sha256(
            ''.join(f'{k}={v}' for k, v in sorted(params.items())).encode()).hexdigest()
        data = {
            **params,
            'apiKey': 'X5BM3w8N7MKozC0B85o4KMlzLZKhV00y',
            'sig': hmac.new(b'acOrvUS15XRW2o9JksiK1KgQ6Vbds8ZW',
                            f'{api_slug}{sha256}'.encode(), hashlib.sha512).hexdigest(),
        }
        return f'{self._DOMAIN}{api_slug}?{urllib.parse.urlencode(data)}'

    def _call_api(self, url_type, params, display_id=None, **kwargs):
        resp = self._download_json(
            self._api_url(url_type, params), display_id or params.get('id'),
            note=f'Downloading {url_type} JSON metadata', **kwargs)
        return (resp or {}).get('data') or {}

    def _real_initialize(self):
        if not self._cookies_passed:
            self._request_webpage(
                self._api_url('bai-hat', {'id': ''}), None, note='Updating cookies')

    def _parse_items(self, items):
        for url in traverse_obj(items, (..., 'link')) or []:
            yield self.url_result(urljoin(self._DOMAIN, url))

    def _fetch_page(self, id_, url_type, page):
        raise NotImplementedError('This method must be implemented by subclasses')

    def _paged_list(self, _id, url_type):
        count = 0
        for page in itertools.count(1):
            data = self._fetch_page(_id, url_type, page)
            entries = list(self._parse_items(data.get('items')))
            count += len(entries)
            yield from entries
            if not data.get('hasMore') or try_call(lambda: count > data['total']):
                break


class ZingMp3IE(ZingMp3BaseIE):
    _VALID_URL = ZingMp3BaseIE._VALID_URL_TMPL % 'bai-hat|video-clip|embed|eps'
    IE_NAME = 'zingmp3'
    IE_DESC = 'zingmp3.vn'
    _TESTS = [{
        'url': 'https://mp3.zing.vn/bai-hat/Xa-Mai-Xa-Bao-Thy/ZWZB9WAB.html',
        'md5': 'ead7ae13693b3205cbc89536a077daed',
        'info_dict': {
            'id': 'ZWZB9WAB',
            'title': 'Xa Mãi Xa',
            'ext': 'mp3',
            'thumbnail': r're:^https?://.+\.jpg',
            'subtitles': {
                'origin': [{
                    'ext': 'lrc',
                }],
            },
            'duration': 255,
            'track': 'Xa Mãi Xa',
            'artist': 'Bảo Thy',
            'album': 'Special Album',
            'album_artist': 'Bảo Thy',
        },
    }, {
        'url': 'https://zingmp3.vn/video-clip/Suong-Hoa-Dua-Loi-K-ICM-RYO/ZO8ZF7C7.html',
        'md5': '92c6e7a019f06b4682a6c35ae5785fab',
        'info_dict': {
            'id': 'ZO8ZF7C7',
            'title': 'Sương Hoa Đưa Lối',
            'ext': 'mp4',
            'thumbnail': r're:^https?://.+\.jpg',
            'duration': 207,
            'track': 'Sương Hoa Đưa Lối',
            'artist': 'K-ICM, RYO',
            'album': 'Sương Hoa Đưa Lối (Single)',
            'album_artist': 'K-ICM, RYO',
        },
    }, {
        'url': 'https://zingmp3.vn/bai-hat/Nguoi-Yeu-Toi-Lanh-Lung-Sat-Da-Mr-Siro/ZZ6IW7OU.html',
        'md5': '3e9f7a9bd0d965573dbff8d7c68b629d',
        'info_dict': {
            'id': 'ZZ6IW7OU',
            'title': 'Người Yêu Tôi Lạnh Lùng Sắt Đá',
            'ext': 'mp3',
            'thumbnail': r're:^https?://.+\.jpg',
            'duration': 303,
            'track': 'Người Yêu Tôi Lạnh Lùng Sắt Đá',
            'artist': 'Mr. Siro',
            'album': 'Người Yêu Tôi Lạnh Lùng Sắt Đá (Single)',
            'album_artist': 'Mr. Siro',
        },
    }, {
        'url': 'https://zingmp3.vn/eps/Cham-x-Ban-Noi-Goi-La-Nha/ZZD9ACWI.html',
        'md5': 'd52f9f63e2631e004e4f15188eedcf80',
        'info_dict': {
            'id': 'ZZD9ACWI',
            'title': 'Chạm x Bạn - Nơi Gọi Là Nhà',
            'ext': 'mp3',
            'duration': 3716,
            'thumbnail': r're:^https?://.+\.jpg',
            'track': 'Chạm x Bạn - Nơi Gọi Là Nhà',
            'artist': 'On Air',
            'album': 'Top Podcast',
            'album_artist': 'On Air',
        },
    }, {
        'url': 'https://zingmp3.vn/embed/song/ZWZEI76B?start=false',
        'only_matching': True,
    }, {
        'url': 'https://zingmp3.vn/bai-hat/Xa-Mai-Xa-Bao-Thy/ZWZB9WAB.html',
        'only_matching': True,
    }]

    def _real_extract(self, url):
        song_id, url_type = self._match_valid_url(url).group('id', 'type')
        item = self._call_api(url_type, {'id': song_id})

        item_id = item.get('encodeId') or song_id
        if url_type == 'video-clip':
            source = item.get('streaming')
            source['mp4'] = self._download_json(
                'http://api.mp3.zing.vn/api/mobile/video/getvideoinfo', item_id,
                query={'requestdata': json.dumps({'id': item_id})},
                note='Downloading mp4 JSON metadata').get('source')
        elif url_type == 'eps':
            source = self._call_api('episode-streaming', {'id': item_id})
        else:
            source = self._call_api('song-streaming', {'id': item_id})

        formats = []
        for k, v in (source or {}).items():
            if not v or v == 'VIP':
                continue
            if k not in ('mp4', 'hls'):
                formats.append({
                    'ext': 'mp3',
                    'format_id': k,
                    'tbr': int_or_none(k),
                    'url': self._proto_relative_url(v),
                    'vcodec': 'none',
                })
                continue
            for res, video_url in v.items():
                if not video_url:
                    continue
                if k == 'hls':
                    formats.extend(self._extract_m3u8_formats(video_url, item_id, 'mp4', m3u8_id=k, fatal=False))
                    continue
                formats.append({
                    'format_id': f'mp4-{res}',
                    'url': video_url,
                    'height': int_or_none(res),
                })

        if not formats:
            if item.get('msg') == 'Sorry, this content is not available in your country.':
                self.raise_geo_restricted(countries=self._GEO_COUNTRIES, metadata_available=True)
            else:
                self.raise_no_formats('The song is only for VIP accounts.')

        lyric = item.get('lyric') or self._call_api('lyric', {'id': item_id}, fatal=False).get('file')

        return {
            'id': item_id,
            'title': traverse_obj(item, 'title', 'alias'),
            'thumbnail': traverse_obj(item, 'thumbnail', 'thumbnailM'),
            'duration': int_or_none(item.get('duration')),
            'track': traverse_obj(item, 'title', 'alias'),
            'artist': traverse_obj(item, 'artistsNames', 'artists_names', ('artists', 0, 'name')),
            'album': traverse_obj(item, ('album', ('name', 'title')), ('genres', 0, 'name'), get_all=False),
            'album_artist': traverse_obj(item, ('album', ('artistsNames', 'artists_names')),
                                         ('artists', 0, 'name'), get_all=False),
            'formats': formats,
            'subtitles': {'origin': [{'url': lyric}]} if lyric else None,
        }


class ZingMp3AlbumIE(ZingMp3BaseIE):
    _VALID_URL = ZingMp3BaseIE._VALID_URL_TMPL % 'album|playlist'
    _TESTS = [{
        'url': 'https://zingmp3.vn/album/Ca-Phe-Quan-Quen-Hoang-Dung-My-Anh-Da-LAB-Thinh-Suy/ZOC7WUZC.html',
        'info_dict': {
            'id': 'ZOC7WUZC',
            'title': 'Cà Phê Quán Quen',
        },
        'playlist_mincount': 10,
    }, {
        'url': 'https://zingmp3.vn/album/Nhung-Bai-Hat-Hay-Nhat-Cua-Mr-Siro-Mr-Siro/ZWZAEZZD.html',
        'info_dict': {
            'id': 'ZWZAEZZD',
            'title': 'Những Bài Hát Hay Nhất Của Mr. Siro',
        },
        'playlist_mincount': 20,
    }, {
        'url': 'http://mp3.zing.vn/playlist/Duong-Hong-Loan-apollobee/IWCAACCB.html',
        'only_matching': True,
    }, {
        'url': 'https://zingmp3.vn/album/Lau-Dai-Tinh-Ai-Bang-Kieu-Minh-Tuyet/ZWZBWDAF.html',
        'only_matching': True,
    }]
    IE_NAME = 'zingmp3:album'

    def _real_extract(self, url):
        song_id, url_type = self._match_valid_url(url).group('id', 'type')
        data = self._call_api(url_type, {'id': song_id})
        return self.playlist_result(
            self._parse_items(traverse_obj(data, ('song', 'items'))),
            traverse_obj(data, 'id', 'encodeId'), traverse_obj(data, 'name', 'title'))


class ZingMp3ChartHomeIE(ZingMp3BaseIE):
    _VALID_URL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<id>(?:zing-chart|moi-phat-hanh|top100|podcast-discover))/?(?:[#?]|$)'
    _TESTS = [{
        'url': 'https://zingmp3.vn/zing-chart',
        'info_dict': {
            'id': 'zing-chart',
        },
        'playlist_mincount': 100,
    }, {
        'url': 'https://zingmp3.vn/moi-phat-hanh',
        'info_dict': {
            'id': 'moi-phat-hanh',
        },
        'playlist_mincount': 100,
    }, {
        'url': 'https://zingmp3.vn/top100',
        'info_dict': {
            'id': 'top100',
        },
        'playlist_mincount': 50,
    }, {
        'url': 'https://zingmp3.vn/podcast-discover',
        'info_dict': {
            'id': 'podcast-discover',
        },
        'playlist_mincount': 4,
    }]
    IE_NAME = 'zingmp3:chart-home'

    def _real_extract(self, url):
        url_type = self._match_id(url)
        params = {'id': url_type}
        if url_type == 'podcast-discover':
            params['type'] = 'discover'
        data = self._call_api(url_type, params)
        items = []
        if url_type == 'top100':
            items.extend(traverse_obj(data, (..., 'items', ..., {dict})))
        elif url_type == 'zing-chart':
            items.extend(traverse_obj(data, ('RTChart', 'items', ..., {dict})))
        else:
            items.extend(traverse_obj(data, ('items', ..., {dict})))
        return self.playlist_result(self._parse_items(items), url_type)


class ZingMp3WeekChartIE(ZingMp3BaseIE):
    _VALID_URL = ZingMp3BaseIE._VALID_URL_TMPL % 'zing-chart-tuan'
    IE_NAME = 'zingmp3:week-chart'
    _TESTS = [{
        'url': 'https://zingmp3.vn/zing-chart-tuan/Bai-hat-Viet-Nam/IWZ9Z08I.html',
        'info_dict': {
            'id': 'IWZ9Z08I',
            'title': 'zing-chart-vn',
        },
        'playlist_mincount': 10,
    }, {
        'url': 'https://zingmp3.vn/zing-chart-tuan/Bai-hat-US-UK/IWZ9Z0BW.html',
        'info_dict': {
            'id': 'IWZ9Z0BW',
            'title': 'zing-chart-us',
        },
        'playlist_mincount': 10,
    }, {
        'url': 'https://zingmp3.vn/zing-chart-tuan/Bai-hat-KPop/IWZ9Z0BO.html',
        'info_dict': {
            'id': 'IWZ9Z0BO',
            'title': 'zing-chart-korea',
        },
        'playlist_mincount': 10,
    }]

    def _real_extract(self, url):
        song_id, url_type = self._match_valid_url(url).group('id', 'type')
        data = self._call_api(url_type, {'id': song_id})
        return self.playlist_result(
            self._parse_items(data['items']), song_id, f'zing-chart-{data.get("country", "")}')


class ZingMp3ChartMusicVideoIE(ZingMp3BaseIE):
    _VALID_URL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<type>the-loai-video)/(?P<regions>[^/]+)/(?P<id>[^\.]+)'
    IE_NAME = 'zingmp3:chart-music-video'
    _TESTS = [{
        'url': 'https://zingmp3.vn/the-loai-video/Viet-Nam/IWZ9Z08I.html',
        'info_dict': {
            'id': 'IWZ9Z08I',
            'title': 'the-loai-video_Viet-Nam',
        },
        'playlist_mincount': 400,
    }, {
        'url': 'https://zingmp3.vn/the-loai-video/Au-My/IWZ9Z08O.html',
        'info_dict': {
            'id': 'IWZ9Z08O',
            'title': 'the-loai-video_Au-My',
        },
        'playlist_mincount': 40,
    }, {
        'url': 'https://zingmp3.vn/the-loai-video/Han-Quoc/IWZ9Z08W.html',
        'info_dict': {
            'id': 'IWZ9Z08W',
            'title': 'the-loai-video_Han-Quoc',
        },
        'playlist_mincount': 30,
    }, {
        'url': 'https://zingmp3.vn/the-loai-video/Khong-Loi/IWZ9Z086.html',
        'info_dict': {
            'id': 'IWZ9Z086',
            'title': 'the-loai-video_Khong-Loi',
        },
        'playlist_mincount': 1,
    }]

    def _fetch_page(self, song_id, url_type, page):
        return self._call_api(url_type, {
            'id': song_id,
            'type': 'genre',
            'page': page,
            'count': self._PER_PAGE,
        })

    def _real_extract(self, url):
        song_id, regions, url_type = self._match_valid_url(url).group('id', 'regions', 'type')
        return self.playlist_result(self._paged_list(song_id, url_type), song_id, f'{url_type}_{regions}')


class ZingMp3UserIE(ZingMp3BaseIE):
    _VALID_URL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<user>[^/]+)/(?P<type>bai-hat|single|album|video|song)/?(?:[?#]|$)'
    IE_NAME = 'zingmp3:user'
    _TESTS = [{
        'url': 'https://zingmp3.vn/Mr-Siro/bai-hat',
        'info_dict': {
            'id': 'IWZ98609',
            'title': 'Mr. Siro - bai-hat',
            'description': 'md5:5bdcf45e955dc1b8d7f518f322ffef36',
        },
        'playlist_mincount': 91,
    }, {
        'url': 'https://zingmp3.vn/Mr-Siro/album',
        'info_dict': {
            'id': 'IWZ98609',
            'title': 'Mr. Siro - album',
            'description': 'md5:5bdcf45e955dc1b8d7f518f322ffef36',
        },
        'playlist_mincount': 3,
    }, {
        'url': 'https://zingmp3.vn/Mr-Siro/single',
        'info_dict': {
            'id': 'IWZ98609',
            'title': 'Mr. Siro - single',
            'description': 'md5:5bdcf45e955dc1b8d7f518f322ffef36',
        },
        'playlist_mincount': 20,
    }, {
        'url': 'https://zingmp3.vn/Mr-Siro/video',
        'info_dict': {
            'id': 'IWZ98609',
            'title': 'Mr. Siro - video',
            'description': 'md5:5bdcf45e955dc1b8d7f518f322ffef36',
        },
        'playlist_mincount': 15,
    }, {
        'url': 'https://zingmp3.vn/new-release/song',
        'info_dict': {
            'id': 'new-release-song',
        },
        'playlist_mincount': 50,
    }, {
        'url': 'https://zingmp3.vn/new-release/album',
        'info_dict': {
            'id': 'new-release-album',
        },
        'playlist_mincount': 20,
    }]

    def _fetch_page(self, user_id, url_type, page):
        url_type = 'user-list-song' if url_type == 'bai-hat' else 'user-list-video'
        return self._call_api(url_type, {
            'id': user_id,
            'type': 'artist',
            'page': page,
            'count': self._PER_PAGE,
        })

    def _real_extract(self, url):
        alias, url_type = self._match_valid_url(url).group('user', 'type')
        if not url_type:
            url_type = 'bai-hat'

        user_info = self._call_api('info-artist', {}, alias, query={'alias': alias})

        # Handle for new-release
        if alias == 'new-release' and url_type in ('song', 'album'):
            _id = f'{alias}-{url_type}'
            return self.playlist_result(self._parse_items(
                self._call_api('new-release', params={'type': url_type}, display_id=_id)), _id)
        else:
            # Handle for user/artist
            if url_type in ('bai-hat', 'video'):
                entries = self._paged_list(user_info['id'], url_type)
            else:
                section_id = 'aAlbum' if url_type == 'album' else 'aSingle'
                entries = self._parse_items(traverse_obj(user_info, (
                    'sections', lambda _, v: v['sectionId'] == section_id, 'items', ...)))
            return self.playlist_result(
                entries, user_info['id'], join_nonempty(user_info.get('name'), url_type, delim=' - '),
                user_info.get('biography'))


class ZingMp3HubIE(ZingMp3BaseIE):
    IE_NAME = 'zingmp3:hub'
    _VALID_URL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<type>hub)/[^/?#]+/(?P<id>[^./?#]+)'
    _TESTS = [{
        'url': 'https://zingmp3.vn/hub/Nhac-Moi/IWZ9Z0CA.html',
        'info_dict': {
            'id': 'IWZ9Z0CA',
            'title': 'BXH Nhạc Mới',
            'description': 'md5:1cc31b68a6f746427b07b2756c22a558',
        },
        'playlist_mincount': 20,
    }, {
        'url': 'https://zingmp3.vn/hub/Nhac-Viet/IWZ9Z087.html',
        'info_dict': {
            'id': 'IWZ9Z087',
            'title': 'Nhạc Việt',
            'description': 'md5:acc976c8bdde64d5c6ee4a92c39f7a77',
        },
        'playlist_mincount': 30,
    }]

    def _real_extract(self, url):
        song_id, url_type = self._match_valid_url(url).group('id', 'type')
        hub_detail = self._call_api(url_type, {'id': song_id})
        entries = self._parse_items(traverse_obj(hub_detail, (
            'sections', lambda _, v: v['sectionId'] == 'hub', 'items', ...)))
        return self.playlist_result(
            entries, song_id, hub_detail.get('title'), hub_detail.get('description'))


class ZingMp3LiveRadioIE(ZingMp3BaseIE):
    IE_NAME = 'zingmp3:liveradio'
    _VALID_URL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<type>(?:liveradio))/(?P<id>\w+)(?:\.html|\?)'
    _TESTS = [{
        'url': 'https://zingmp3.vn/liveradio/IWZ979UB.html',
        'info_dict': {
            'id': 'IWZ979UB',
            'title': r're:^V\-POP',
            'description': 'md5:aa857f8a91dc9ce69e862a809e4bdc10',
            'ext': 'mp4',
            'view_count': int,
            'thumbnail': r're:^https?://.*\.jpg',
            'like_count': int,
            'live_status': 'is_live',
        },
        'params': {
            'skip_download': True,
        },
    }, {
        'url': 'https://zingmp3.vn/liveradio/IWZ97CWB.html',
        'info_dict': {
            'id': 'IWZ97CWB',
            'title': r're:^Live\s247',
            'description': 'md5:d41d8cd98f00b204e9800998ecf8427e',
            'ext': 'm4a',
            'view_count': int,
            'thumbnail': r're:^https?://.*\.jpg',
            'like_count': int,
            'live_status': 'is_live',
        },
        'params': {
            'skip_download': True,
        },
    }]

    def _real_extract(self, url):
        url_type, live_radio_id = self._match_valid_url(url).group('type', 'id')
        info = self._call_api(url_type, {'id': live_radio_id})
        manifest_url = info.get('streaming')
        if not manifest_url:
            raise ExtractorError('This radio is offline.', expected=True)
        fmts, subtitles = self._extract_m3u8_formats_and_subtitles(manifest_url, live_radio_id, fatal=False)
        return {
            'id': live_radio_id,
            'is_live': True,
            'formats': fmts,
            'subtitles': subtitles,
            **traverse_obj(info, {
                'title': 'title',
                'thumbnail': (('thumbnail', 'thumbnailM', 'thumbnailV', 'thumbnailH'), {url_or_none}),
                'view_count': ('activeUsers', {int_or_none}),
                'like_count': ('totalReaction', {int_or_none}),
                'description': 'description',
            }, get_all=False),
        }


class ZingMp3PodcastEpisodeIE(ZingMp3BaseIE):
    IE_NAME = 'zingmp3:podcast-episode'
    _VALID_URL = ZingMp3BaseIE._VALID_URL_TMPL % 'pgr|cgr'
    _TESTS = [{
        'url': 'https://zingmp3.vn/pgr/Nhac-Moi-Moi-Ngay/68Z9W66B.html',
        'info_dict': {
            'id': '68Z9W66B',
            'title': 'Nhạc Mới Mỗi Ngày',
            'description': 'md5:2875dfa951f8e5356742f1610cf20691',
        },
        'playlist_mincount': 20,
    }, {
        'url': 'https://zingmp3.vn/cgr/Am-nhac/IWZ980AO.html',
        'info_dict': {
            'id': 'IWZ980AO',
            'title': 'Âm nhạc',
        },
        'playlist_mincount': 2,
    }]

    def _fetch_page(self, eps_id, url_type, page):
        return self._call_api(url_type, {
            'id': eps_id,
            'page': page,
            'count': self._PER_PAGE,
        })

    def _real_extract(self, url):
        podcast_id, url_type = self._match_valid_url(url).group('id', 'type')
        podcast_info = self._call_api(url_type, {'id': podcast_id})
        entries = self._paged_list(podcast_id, 'pgr-list' if url_type == 'pgr' else 'cgr-list')
        return self.playlist_result(
            entries, podcast_id, podcast_info.get('title'), podcast_info.get('description'))


class ZingMp3PodcastIE(ZingMp3BaseIE):
    IE_NAME = 'zingmp3:podcast'
    _VALID_URL = r'https?://(?:mp3\.zing|zingmp3)\.vn/(?P<id>(?:cgr|top-podcast|podcast-new))/?(?:[#?]|$)'
    _TESTS = [{
        'url': 'https://zingmp3.vn/cgr',
        'info_dict': {
            'id': 'cgr',
        },
        'playlist_mincount': 5,
    }, {
        'url': 'https://zingmp3.vn/top-podcast',
        'info_dict': {
            'id': 'top-podcast',
        },
        'playlist_mincount': 7,
    }, {
        'url': 'https://zingmp3.vn/podcast-new',
        'info_dict': {
            'id': 'podcast-new',
        },
        'playlist_mincount': 4,
    }]

    def _real_extract(self, url):
        url_type = self._match_id(url)
        params = {'id': url_type}
        if url_type == 'podcast-new':
            params['type'] = 'new'
        items = self._call_api('cgrs' if url_type == 'cgr' else url_type, params)['items']
        return self.playlist_result(self._parse_items(items), url_type)



from .art19 import Art19IE
from .common import InfoExtractor
from ..utils import extract_attributes, get_elements_html_by_class
from ..utils.traversal import traverse_obj


class RideHomeIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?ridehome\.info/show/[\w-]+/(?P<id>[\w-]+)/?(?:$|[?#])'
    _TESTS = [{
        'url': 'https://www.ridehome.info/show/techmeme-ride-home/thu-1228-will-2024-be-the-year-apple-gets-serious-about-gaming-on-macs/',
        'info_dict': {
            'id': 'thu-1228-will-2024-be-the-year-apple-gets-serious-about-gaming-on-macs',
        },
        'playlist_count': 1,
        'playlist': [{
            'md5': 'c84ea3cc96950a9ab86fe540f3edc588',
            'info_dict': {
                'id': '540e5493-9fe6-4c14-a488-dc508d8794b2',
                'ext': 'mp3',
                'title': 'Thu. 12/28 – Will 2024 Be The Year Apple Gets Serious About Gaming On Macs?',
                'description': 'md5:9dba86ae9b5047a8150eceddeeb629c2',
                'series': 'Techmeme Ride Home',
                'series_id': '3c30e8f4-ab48-415b-9421-1ae06cd4058b',
                'upload_date': '20231228',
                'timestamp': 1703780995,
                'modified_date': '20231230',
                'episode_id': '540e5493-9fe6-4c14-a488-dc508d8794b2',
                'modified_timestamp': 1703912404,
                'release_date': '20231228',
                'release_timestamp': 1703782800,
                'duration': 1000.1502,
                'thumbnail': r're:^https?://content\.production\.cdn\.art19\.com/images/.*\.jpeg$',
            },
        }],
    }, {
        'url': 'https://www.ridehome.info/show/techmeme-ride-home/portfolio-profile-sensel-with-ilyarosenberg/',
        'info_dict': {
            'id': 'portfolio-profile-sensel-with-ilyarosenberg',
        },
        'playlist_count': 1,
        'playlist': [{
            'md5': 'bf9d6efad221008ce71aea09d5533cf6',
            'info_dict': {
                'id': '6beed803-b1ef-4536-9fef-c23cf6b4dcac',
                'ext': 'mp3',
                'title': '(Portfolio Profile) Sensel - With @IlyaRosenberg',
                'description': 'md5:e1e4a970bce04290e0ba6f030b0125db',
                'series': 'Techmeme Ride Home',
                'series_id': '3c30e8f4-ab48-415b-9421-1ae06cd4058b',
                'upload_date': '20220108',
                'timestamp': 1641656064,
                'modified_date': '20230418',
                'episode_id': '6beed803-b1ef-4536-9fef-c23cf6b4dcac',
                'modified_timestamp': 1681843318,
                'release_date': '20220108',
                'release_timestamp': 1641672000,
                'duration': 2789.38122,
                'thumbnail': r're:^https?://content\.production\.cdn\.art19\.com/images/.*\.jpeg$',
            },
        }],
    }, {
        'url': 'https://www.ridehome.info/show/spacecasts/big-tech-news-apples-macbook-pro-event/',
        'info_dict': {
            'id': 'big-tech-news-apples-macbook-pro-event',
        },
        'playlist_count': 1,
        'playlist': [{
            'md5': 'b1428530c6e03904a8271e978007fc05',
            'info_dict': {
                'id': 'f4780044-6c4b-4ce0-8215-8a86cc66bff7',
                'ext': 'mp3',
                'title': 'md5:e6c05d44d59b6577a4145ac339de5040',
                'description': 'md5:14152f7228c8a301a77e3d6bc891b145',
                'series': 'SpaceCasts',
                'series_id': '8e3e837d-7fe0-4a23-8e11-894917e07e17',
                'upload_date': '20211026',
                'timestamp': 1635271450,
                'modified_date': '20230502',
                'episode_id': 'f4780044-6c4b-4ce0-8215-8a86cc66bff7',
                'modified_timestamp': 1683057500,
                'release_date': '20211026',
                'release_timestamp': 1635272124,
                'duration': 2266.30531,
                'thumbnail': r're:^https?://content\.production\.cdn\.art19\.com/images/.*\.jpeg$',
            },
        }],
    }]

    def _real_extract(self, url):
        article_id = self._match_id(url)
        webpage = self._download_webpage(url, article_id)

        urls = traverse_obj(
            get_elements_html_by_class('iframeContainer', webpage),
            (..., {extract_attributes}, lambda k, v: k == 'data-src' and Art19IE.suitable(v)))
        return self.playlist_from_matches(urls, article_id, ie=Art19IE)



from .common import InfoExtractor
from .jwplatform import JWPlatformIE


class BundesligaIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?bundesliga\.com/[a-z]{2}/bundesliga/videos(?:/[^?]+)?\?vid=(?P<id>[a-zA-Z0-9]{8})'
    _TESTS = [
        {
            'url': 'https://www.bundesliga.com/en/bundesliga/videos?vid=bhhHkKyN',
            'md5': '8fc3b25cd12440e3a8cdc51f1493849c',
            'info_dict': {
                'id': 'bhhHkKyN',
                'ext': 'mp4',
                'title': 'Watch: Alphonso Davies and Jeremie Frimpong head-to-head',
                'thumbnail': 'https://cdn.jwplayer.com/v2/media/bhhHkKyN/poster.jpg?width=720',
                'upload_date': '20220928',
                'duration': 146,
                'timestamp': 1664366511,
                'description': 'md5:803d4411bd134140c774021dd4b7598b',
            },
        },
        {
            'url': 'https://www.bundesliga.com/en/bundesliga/videos/latest-features/T8IKc8TX?vid=ROHjs06G',
            'only_matching': True,
        },
        {
            'url': 'https://www.bundesliga.com/en/bundesliga/videos/goals?vid=mOG56vWA',
            'only_matching': True,
        },
    ]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        return self.url_result(f'jwplatform:{video_id}', JWPlatformIE, video_id)



import base64
import json

from .common import InfoExtractor
from ..networking.exceptions import HTTPError
from ..utils import (
    ExtractorError,
    clean_html,
    int_or_none,
    parse_qs,
    str_or_none,
    strftime_or_none,
    update_url,
    update_url_query,
    url_or_none,
)
from ..utils.traversal import traverse_obj


class OnsenIE(InfoExtractor):
    IE_NAME = 'onsen'
    IE_DESC = 'インターネットラジオステーション＜音泉＞'

    _API_HEADERS = {'X-Client': 'onsen-web'}
    _BASE_URL = 'https://www.onsen.ag'
    _HEADERS = {'Referer': f'{_BASE_URL}/'}
    _NETRC_MACHINE = 'onsen'
    _VALID_URL = r'https?://(?:(?:share|www)\.)onsen\.ag/program/(?P<id>[^/?#]+)'
    _TESTS = [{
        'url': 'https://share.onsen.ag/program/onsenking?p=90&c=MTA0NjI',
        'info_dict': {
            'id': '10462',
            'ext': 'm4a',
            'title': '第SP回',
            'cast': 'count:3',
            'description': 'md5:de62c80a41c4c8d84da53a1ee681ad18',
            'display_id': 'MTA0NjI=',
            'media_type': 'sound',
            'section_start': 0,
            'series': '音泉キング「下野紘」のラジオ きみはもちろん、＜音泉＞ファミリーだよね？',
            'series_id': 'onsenking',
            'tags': 'count:2',
            'thumbnail': r're:https?://d3bzklg4lms4gh\.cloudfront\.net/program_info/image/default/production/.+',
            'upload_date': '20220627',
            'webpage_url': 'https://www.onsen.ag/program/onsenking?c=MTA0NjI=',
        },
    }, {
        'url': 'https://share.onsen.ag/program/girls-band-cry-radio?p=370&c=MTgwMDE',
        'info_dict': {
            'id': '18001',
            'ext': 'mp4',
            'title': '第4回',
            'cast': 'count:5',
            'description': 'md5:bbca8a389d99c90cbbce8f383c85fedd',
            'display_id': 'MTgwMDE=',
            'media_type': 'movie',
            'section_start': 0,
            'series': 'TVアニメ『ガールズバンドクライ』WEBラジオ「ガールズバンドクライ～ラジオにも全部ぶち込め。～」',
            'series_id': 'girls-band-cry-radio',
            'tags': 'count:3',
            'thumbnail': r're:https?://d3bzklg4lms4gh\.cloudfront\.net/program_info/image/default/production/.+',
            'upload_date': '20240425',
            'webpage_url': 'https://www.onsen.ag/program/girls-band-cry-radio?c=MTgwMDE=',
        },
        'skip': 'Only available for premium supporters',
    }, {
        'url': 'https://www.onsen.ag/program/uma',
        'info_dict': {
            'id': 'uma',
            'title': 'UMA YELL RADIO',
        },
        'playlist_mincount': 35,
    }]

    @classmethod
    def get_temp_id(cls, url):
        import base64
        import urllib.parse

        if c := urllib.parse.parse_qs(urllib.parse.urlparse(url).query).get('c'):
            return base64.urlsafe_b64decode(f'{c[-1]}===').decode()
        return super().get_temp_id(url)

    @staticmethod
    def _get_encoded_id(program):
        return base64.urlsafe_b64encode(str(program['id']).encode()).decode()

    def _perform_login(self, username, password):
        sign_in = self._download_json(
            f'{self._BASE_URL}/web_api/signin', None, 'Logging in', headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                **self._API_HEADERS,
            }, data=json.dumps({
                'session': {
                    'email': username,
                    'password': password,
                },
            }).encode(), expected_status=401)

        if sign_in.get('error'):
            raise ExtractorError('Invalid username or password', expected=True)

    def _real_extract(self, url):
        program_id = self._match_id(url)
        try:
            programs = self._download_json(
                f'{self._BASE_URL}/web_api/programs/{program_id}',
                program_id, headers=self._API_HEADERS)
        except ExtractorError as e:
            if isinstance(e.cause, HTTPError) and e.cause.status == 404:
                raise ExtractorError('Invalid URL', expected=True)
            raise

        query = {k: v[-1] for k, v in parse_qs(url).items() if v}
        if 'c' not in query:
            entries = [
                self.url_result(
                    update_url_query(url, {'c': self._get_encoded_id(program)}),
                    OnsenIE, str_or_none(program['id']),
                ) for program in traverse_obj(programs, ('contents', lambda _, v: v['id']))
            ]

            return self.playlist_result(
                entries, program_id, traverse_obj(programs, ('program_info', 'title', {clean_html})))

        raw_id = base64.urlsafe_b64decode(f'{query["c"]}===').decode()
        p_keys = ('contents', lambda _, v: v['id'] == int(raw_id))

        program = traverse_obj(programs, (*p_keys, any))
        if not program:
            raise ExtractorError(
                'This program is no longer available', expected=True)
        m3u8_url = traverse_obj(program, ('streaming_url', {url_or_none}))
        if not m3u8_url:
            self.raise_login_required(
                'This program is only available for premium supporters')

        display_id = self._get_encoded_id(program)
        date_str = self._search_regex(
            rf'{program_id}0?(\d{{6}})', m3u8_url, 'date string', default=None)

        return {
            'display_id': display_id,
            'formats': self._extract_m3u8_formats(m3u8_url, raw_id, headers=self._HEADERS),
            'http_headers': self._HEADERS,
            'section_start': int_or_none(query.get('t', 0)),
            'upload_date': strftime_or_none(f'20{date_str}'),
            'webpage_url': f'{self._BASE_URL}/program/{program_id}?c={display_id}',
            **traverse_obj(program, {
                'id': ('id', {int}, {str_or_none}),
                'title': ('title', {clean_html}),
                'media_type': ('media_type', {str}),
                'thumbnail': ('poster_image_url', {url_or_none}, {update_url(query=None)}),
            }),
            **traverse_obj(programs, {
                'cast': (('performers', (*p_keys, 'guests')), ..., 'name', {str}, filter),
                'series_id': ('directory_name', {str}),
            }),
            **traverse_obj(programs, ('program_info', {
                'description': ('description', {clean_html}, filter),
                'series': ('title', {clean_html}),
                'tags': ('hashtag_list', ..., {str}, filter),
            })),
        }



from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    extract_attributes,
    get_elements_html_by_class,
    url_or_none,
)
from ..utils.traversal import traverse_obj


class GBNewsIE(InfoExtractor):
    IE_DESC = 'GB News clips, features and live streams'
    _VALID_URL = r'https?://(?:www\.)?gbnews\.(?:uk|com)/(?:\w+/)?(?P<id>[^#?]+)'

    _PLATFORM = 'safari'
    _SSMP_URL = 'https://mm-v2.simplestream.com/ssmp/api.php'
    _TESTS = [{
        'url': 'https://www.gbnews.com/news/bbc-claudine-gay-harvard-university-antisemitism-row',
        'info_dict': {
            'id': '52264136',
            'ext': 'mp4',
            'thumbnail': r're:https?://www\.gbnews\.\w+/.+\.(?:jpe?g|png|webp)',
            'display_id': 'bbc-claudine-gay-harvard-university-antisemitism-row',
            'description': 'The post was criticised by former employers of the broadcaster',
            'title': 'BBC deletes post after furious backlash over headline downplaying antisemitism',
        },
    }, {
        'url': 'https://www.gbnews.com/royal/prince-harry-in-love-with-kate-meghan-markle-jealous-royal',
        'info_dict': {
            'id': '52328390',
            'ext': 'mp4',
            'thumbnail': r're:https?://www\.gbnews\.\w+/.+\.(?:jpe?g|png|webp)',
            'display_id': 'prince-harry-in-love-with-kate-meghan-markle-jealous-royal',
            'description': 'Ingrid Seward has published 17 books documenting the highs and lows of the Royal Family',
            'title': 'Royal author claims Prince Harry was \'in love\' with Kate - Meghan was \'jealous\'',
        },
    }, {
        'url': 'https://www.gbnews.uk/watchlive',
        'info_dict': {
            'id': '1069',
            'ext': 'mp4',
            'thumbnail': r're:https?://www\.gbnews\.\w+/.+\.(?:jpe?g|png|webp)',
            'display_id': 'watchlive',
            'live_status': 'is_live',
            'title': r're:^GB News Live',
        },
        'params': {'skip_download': 'm3u8'},
    }]
    _SS_ENDPOINTS = None

    def _get_ss_endpoint(self, data_id, data_env):
        if not self._SS_ENDPOINTS:
            self._SS_ENDPOINTS = {}

        if not data_id:
            data_id = 'GB003'
        if not data_env:
            data_env = 'production'
        key = data_id, data_env
        result = self._SS_ENDPOINTS.get(key)
        if result:
            return result

        json_data = self._download_json(
            self._SSMP_URL, None, 'Downloading Simplestream JSON metadata', query={
                'id': data_id,
                'env': data_env,
            })
        meta_url = traverse_obj(json_data, ('response', 'api_hostname', {url_or_none}))
        if not meta_url:
            raise ExtractorError('No API host found')

        self._SS_ENDPOINTS[key] = meta_url
        return meta_url

    def _real_extract(self, url):
        display_id = self._match_id(url).rpartition('/')[2]
        webpage = self._download_webpage(url, display_id)

        video_data = None
        elements = get_elements_html_by_class('simplestream', webpage)
        for html_tag in elements:
            attributes = extract_attributes(html_tag)
            if 'sidebar' not in (attributes.get('class') or ''):
                video_data = attributes
        if not video_data:
            raise ExtractorError('Could not find video element', expected=True)

        endpoint_url = self._get_ss_endpoint(video_data.get('data-id'), video_data.get('data-env'))

        uvid = video_data['data-uvid']
        video_type = video_data.get('data-type')
        if not video_type or video_type == 'vod':
            video_type = 'show'
        stream_data = self._download_json(
            f'{endpoint_url}/api/{video_type}/stream/{uvid}',
            uvid, 'Downloading stream JSON', query={
                'key': video_data.get('data-key'),
                'platform': self._PLATFORM,
            })
        if traverse_obj(stream_data, 'drm'):
            self.report_drm(uvid)

        return {
            'id': uvid,
            'display_id': display_id,
            'title': self._og_search_title(webpage, default=None),
            'description': self._og_search_description(webpage, default=None),
            'formats': self._extract_m3u8_formats(traverse_obj(stream_data, (
                'response', 'stream', {url_or_none})), uvid, 'mp4'),
            'thumbnail': self._og_search_thumbnail(webpage, default=None),
            'is_live': video_type == 'live',
        }



import json
import urllib.parse

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    make_archive_id,
    unified_timestamp,
    urljoin,
)
from ..utils.traversal import traverse_obj


class NintendoIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?nintendo\.com/(?:(?P<locale>\w{2}(?:-\w{2})?)/)?nintendo-direct/(?P<slug>[^/?#]+)'
    _TESTS = [{
        'url': 'https://www.nintendo.com/nintendo-direct/09-04-2019/',
        'info_dict': {
            'ext': 'mp4',
            'id': '2oPmiviVePUA1IqAZzjuVh',
            'display_id': '09-04-2019',
            'title': 'Nintendo Direct 9.4.2019',
            'timestamp': 1567580400,
            'description': 'md5:8aac2780361d8cb772b6d1de66d7d6f4',
            'upload_date': '20190904',
            'age_limit': 17,
            '_old_archive_ids': ['nintendo J2bXdmaTE6fe3dWJTPcc7m23FNbc_A1V'],
        },
    }, {
        'url': 'https://www.nintendo.com/en-ca/nintendo-direct/08-31-2023/',
        'info_dict': {
            'ext': 'mp4',
            'id': '2TB2w2rJhNYF84qQ9E57hU',
            'display_id': '08-31-2023',
            'title': 'Super Mario Bros. Wonder Direct 8.31.2023',
            'timestamp': 1693465200,
            'description': 'md5:3067c5b824bcfdae9090a7f38ab2d200',
            'tags': ['Mild Fantasy Violence', 'In-Game Purchases'],
            'upload_date': '20230831',
            'age_limit': 6,
        },
    }, {
        'url': 'https://www.nintendo.com/us/nintendo-direct/50-fact-extravaganza/',
        'info_dict': {
            'ext': 'mp4',
            'id': 'j0BBGzfw0pQ',
            'channel_follower_count': int,
            'view_count': int,
            'description': 'Learn new details about Super Smash Bros. for Wii U, which launches on November 21.',
            'duration': 2123,
            'availability': 'public',
            'thumbnail': 'https://i.ytimg.com/vi_webp/j0BBGzfw0pQ/maxresdefault.webp',
            'timestamp': 1414047600,
            'channel_id': 'UCGIY_O-8vW4rfX98KlMkvRg',
            'chapters': 'count:53',
            'heatmap': 'count:100',
            'upload_date': '20141023',
            'uploader_id': '@NintendoAmerica',
            'playable_in_embed': True,
            'categories': ['Gaming'],
            'display_id': '50-fact-extravaganza',
            'channel': 'Nintendo of America',
            'tags': ['Comic Mischief', 'Cartoon Violence', 'Mild Suggestive Themes'],
            'like_count': int,
            'channel_url': 'https://www.youtube.com/channel/UCGIY_O-8vW4rfX98KlMkvRg',
            'age_limit': 10,
            'uploader_url': 'https://www.youtube.com/@NintendoAmerica',
            'comment_count': int,
            'live_status': 'not_live',
            'uploader': 'Nintendo of America',
            'title': '50-FACT Extravaganza',
        },
    }]

    def _create_asset_url(self, path):
        return urljoin('https://assets.nintendo.com/', urllib.parse.quote(path))

    def _real_extract(self, url):
        locale, slug = self._match_valid_url(url).group('locale', 'slug')

        language, _, country = (locale or 'US').rpartition('-')
        parsed_locale = f'{language.lower() or "en"}_{country.upper()}'
        self.write_debug(f'Using locale {parsed_locale} (from {locale})', only_once=True)

        response = self._download_json('https://graph.nintendo.com/', slug, query={
            'operationName': 'NintendoDirect',
            'variables': json.dumps({
                'locale': parsed_locale,
                'slug': slug,
            }, separators=(',', ':')),
            'extensions': json.dumps({
                'persistedQuery': {
                    'version': 1,
                    'sha256Hash': '969b16fe9f08b686fa37bc44d1fd913b6188e65794bb5e341c54fa683a8004cb',
                },
            }, separators=(',', ':')),
        })
        # API returns `{"data": {"direct": null}}` if no matching id
        direct_info = traverse_obj(response, ('data', 'direct', {dict}))
        if not direct_info:
            raise ExtractorError(f'No Nintendo Direct with id {slug} exists', expected=True)

        errors = ', '.join(traverse_obj(response, ('errors', ..., 'message')))
        if errors:
            raise ExtractorError(f'GraphQL API error: {errors or "Unknown error"}')

        result = traverse_obj(direct_info, {
            'id': ('id', {str}),
            'title': ('name', {str}),
            'timestamp': ('startDate', {unified_timestamp}),
            'description': ('description', 'text', {str}),
            'age_limit': ('contentRating', 'order', {int}),
            'tags': ('contentDescriptors', ..., 'label', {str}),
            'thumbnail': ('thumbnail', {self._create_asset_url}),
        })
        result['display_id'] = slug

        asset_id = traverse_obj(direct_info, ('video', 'publicId', {str}))
        if not asset_id:
            youtube_id = traverse_obj(direct_info, ('liveStream', {str}))
            if not youtube_id:
                self.raise_no_formats('Could not find any video formats', video_id=slug)

            return self.url_result(youtube_id, **result, url_transparent=True)

        if asset_id.startswith('Legacy Videos/'):
            result['_old_archive_ids'] = [make_archive_id(self, asset_id[14:])]
        result['formats'] = self._extract_m3u8_formats(
            self._create_asset_url(f'/video/upload/sp_full_hd/v1/{asset_id}.m3u8'), slug)

        return result



import functools

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    OnDemandPagedList,
    clean_html,
    determine_ext,
    float_or_none,
    format_field,
    int_or_none,
    join_nonempty,
    parse_iso8601,
    url_or_none,
    urljoin,
)
from ..utils.traversal import require, traverse_obj


class FloatplaneBaseIE(InfoExtractor):
    def _real_extract(self, url):
        post_id = self._match_id(url)

        post_data = self._download_json(
            f'{self._BASE_URL}/api/v3/content/post', post_id, query={'id': post_id},
            note='Downloading post data', errnote='Unable to download post data',
            impersonate=self._IMPERSONATE_TARGET)

        if not any(traverse_obj(post_data, ('metadata', ('hasVideo', 'hasAudio')))):
            raise ExtractorError('Post does not contain a video or audio track', expected=True)

        uploader_url = format_field(
            post_data, [('creator', 'urlname')], f'{self._BASE_URL}/channel/%s/home') or None

        common_info = {
            'uploader_url': uploader_url,
            'channel_url': urljoin(f'{uploader_url}/', traverse_obj(post_data, ('channel', 'urlname'))),
            'availability': self._availability(needs_subscription=True),
            **traverse_obj(post_data, {
                'uploader': ('creator', 'title', {str}),
                'uploader_id': ('creator', 'id', {str}),
                'channel': ('channel', 'title', {str}),
                'channel_id': ('channel', 'id', {str}),
                'release_timestamp': ('releaseDate', {parse_iso8601}),
            }),
        }

        items = []
        for media in traverse_obj(post_data, (('videoAttachments', 'audioAttachments'), ...)):
            media_id = media['id']
            media_typ = media.get('type') or 'video'

            stream = self._download_json(
                f'{self._BASE_URL}/api/v3/delivery/info', media_id,
                query={'scenario': 'onDemand', 'entityId': media_id},
                note=f'Downloading {media_typ} stream data',
                impersonate=self._IMPERSONATE_TARGET)

            metadata = self._download_json(
                f'{self._BASE_URL}/api/v3/content/{media_typ}', media_id,
                f'Downloading {media_typ} metadata', query={'id': media_id},
                fatal=False, impersonate=self._IMPERSONATE_TARGET)

            cdn_base_url = traverse_obj(stream, (
                'groups', 0, 'origins', ..., 'url', {url_or_none}, any, {require('cdn base url')}))

            formats = []
            for variant in traverse_obj(stream, ('groups', 0, 'variants', lambda _, v: v['url'])):
                format_url = urljoin(cdn_base_url, variant['url'])
                format_id = traverse_obj(variant, ('name', {str}))
                hls_aes = {}
                m3u8_data = None

                # If we need impersonation for the API, then we need it for HLS keys too: extract in advance
                if self._IMPERSONATE_TARGET is not None:
                    m3u8_data = self._download_webpage(
                        format_url, media_id, fatal=False, impersonate=self._IMPERSONATE_TARGET, headers=self._HEADERS,
                        note=join_nonempty('Downloading', format_id, 'm3u8 information', delim=' '),
                        errnote=join_nonempty('Failed to download', format_id, 'm3u8 information', delim=' '))
                    if not m3u8_data:
                        continue

                    key_url = self._search_regex(
                        r'#EXT-X-KEY:METHOD=AES-128,URI="(https?://[^"]+)"',
                        m3u8_data, 'HLS AES key URI', default=None)
                    if key_url:
                        urlh = self._request_webpage(
                            key_url, media_id, fatal=False, impersonate=self._IMPERSONATE_TARGET, headers=self._HEADERS,
                            note=join_nonempty('Downloading', format_id, 'HLS AES key', delim=' '),
                            errnote=join_nonempty('Failed to download', format_id, 'HLS AES key', delim=' '))
                        if urlh:
                            hls_aes['key'] = urlh.read().hex()

                formats.append({
                    **traverse_obj(variant, {
                        'format_note': ('label', {str}),
                        'width': ('meta', 'video', 'width', {int_or_none}),
                        'height': ('meta', 'video', 'height', {int_or_none}),
                        'vcodec': ('meta', 'video', 'codec', {str}),
                        'acodec': ('meta', 'audio', 'codec', {str}),
                        'vbr': ('meta', 'video', 'bitrate', 'average', {int_or_none(scale=1000)}),
                        'abr': ('meta', 'audio', 'bitrate', 'average', {int_or_none(scale=1000)}),
                        'audio_channels': ('meta', 'audio', 'channelCount', {int_or_none}),
                        'fps': ('meta', 'video', 'fps', {float_or_none}),
                    }),
                    'url': format_url,
                    'ext': determine_ext(format_url.partition('/chunk.m3u8')[0], 'mp4'),
                    'format_id': format_id,
                    'hls_media_playlist_data': m3u8_data,
                    'hls_aes': hls_aes or None,
                })

            subtitles = {}
            automatic_captions = {}
            for sub_data in traverse_obj(metadata, ('textTracks', lambda _, v: url_or_none(v['src']))):
                sub_lang = sub_data.get('language') or 'en'
                sub_entry = {'url': sub_data['src']}
                if sub_data.get('generated'):
                    automatic_captions.setdefault(sub_lang, []).append(sub_entry)
                else:
                    subtitles.setdefault(sub_lang, []).append(sub_entry)

            items.append({
                **common_info,
                'id': media_id,
                **traverse_obj(metadata, {
                    'title': ('title', {str}),
                    'duration': ('duration', {int_or_none}),
                    'thumbnail': ('thumbnail', 'path', {url_or_none}),
                }),
                'formats': formats,
                'subtitles': subtitles,
                'automatic_captions': automatic_captions,
            })

        post_info = {
            **common_info,
            'id': post_id,
            'display_id': post_id,
            **traverse_obj(post_data, {
                'title': ('title', {str}),
                'description': ('text', {clean_html}),
                'like_count': ('likes', {int_or_none}),
                'dislike_count': ('dislikes', {int_or_none}),
                'comment_count': ('comments', {int_or_none}),
                'thumbnail': ('thumbnail', 'path', {url_or_none}),
            }),
            'http_headers': self._HEADERS,
        }

        if len(items) > 1:
            return self.playlist_result(items, **post_info)

        post_info.update(items[0])
        return post_info


class FloatplaneIE(FloatplaneBaseIE):
    _VALID_URL = r'https?://(?:(?:www|beta)\.)?floatplane\.com/post/(?P<id>\w+)'
    _BASE_URL = 'https://www.floatplane.com'
    _IMPERSONATE_TARGET = None
    _HEADERS = {
        'Origin': _BASE_URL,
        'Referer': f'{_BASE_URL}/',
    }
    _TESTS = [{
        'url': 'https://www.floatplane.com/post/2Yf3UedF7C',
        'info_dict': {
            'id': 'yuleLogLTT',
            'ext': 'mp4',
            'display_id': '2Yf3UedF7C',
            'title': '8K Yule Log Fireplace with Crackling Fire Sounds - 10 Hours',
            'description': 'md5:adf2970e0de1c5e3df447818bb0309f6',
            'thumbnail': r're:^https?://.*\.jpe?g$',
            'duration': 36035,
            'comment_count': int,
            'like_count': int,
            'dislike_count': int,
            'release_date': '20191206',
            'release_timestamp': 1575657000,
            'uploader': 'LinusTechTips',
            'uploader_id': '59f94c0bdd241b70349eb72b',
            'uploader_url': 'https://www.floatplane.com/channel/linustechtips/home',
            'channel': 'Linus Tech Tips',
            'channel_id': '63fe42c309e691e4e36de93d',
            'channel_url': 'https://www.floatplane.com/channel/linustechtips/home/main',
            'availability': 'subscriber_only',
        },
        'params': {'skip_download': 'm3u8'},
    }, {
        'url': 'https://www.floatplane.com/post/j2jqG3JmgJ',
        'info_dict': {
            'id': 'j2jqG3JmgJ',
            'title': 'TJM: Does Anyone Care About Avatar: The Way of Water?',
            'description': 'md5:00bf17dc5733e4031e99b7fd6489f274',
            'thumbnail': r're:^https?://.*\.jpe?g$',
            'comment_count': int,
            'like_count': int,
            'dislike_count': int,
            'release_timestamp': 1671915900,
            'release_date': '20221224',
            'uploader': 'LinusTechTips',
            'uploader_id': '59f94c0bdd241b70349eb72b',
            'uploader_url': 'https://www.floatplane.com/channel/linustechtips/home',
            'channel': "They're Just Movies",
            'channel_id': '64135f82fc76ab7f9fbdc876',
            'channel_url': 'https://www.floatplane.com/channel/linustechtips/home/tajm',
            'availability': 'subscriber_only',
        },
        'playlist_count': 2,
    }, {
        'url': 'https://www.floatplane.com/post/3tK2tInhoN',
        'info_dict': {
            'id': '3tK2tInhoN',
            'title': 'Extras - How Linus Communicates with Editors (Compensator 4)',
            'description': 'md5:83cd40aae1ce124df33769600c80ca5b',
            'thumbnail': r're:^https?://.*\.jpe?g$',
            'comment_count': int,
            'like_count': int,
            'dislike_count': int,
            'release_timestamp': 1700529120,
            'release_date': '20231121',
            'uploader': 'LinusTechTips',
            'uploader_id': '59f94c0bdd241b70349eb72b',
            'uploader_url': 'https://www.floatplane.com/channel/linustechtips/home',
            'channel': 'FP Exclusives',
            'channel_id': '6413623f5b12cca228a28e78',
            'channel_url': 'https://www.floatplane.com/channel/linustechtips/home/fpexclusive',
            'availability': 'subscriber_only',
        },
        'playlist_count': 2,
    }, {
        'url': 'https://beta.floatplane.com/post/d870PEFXS1',
        'info_dict': {
            'id': 'bg9SuYKEww',
            'ext': 'mp4',
            'display_id': 'd870PEFXS1',
            'title': 'LCS Drama, TLOU 2 Remaster, Destiny 2 Player Count Drops, + More!',
            'description': 'md5:80d612dcabf41b17487afcbe303ec57d',
            'thumbnail': r're:^https?://.*\.jpe?g$',
            'release_timestamp': 1700622000,
            'release_date': '20231122',
            'duration': 513,
            'like_count': int,
            'dislike_count': int,
            'comment_count': int,
            'uploader': 'LinusTechTips',
            'uploader_id': '59f94c0bdd241b70349eb72b',
            'uploader_url': 'https://www.floatplane.com/channel/linustechtips/home',
            'channel': 'GameLinked',
            'channel_id': '649dbade3540dbc3945eeda7',
            'channel_url': 'https://www.floatplane.com/channel/linustechtips/home/gamelinked',
            'availability': 'subscriber_only',
        },
        'params': {'skip_download': 'm3u8'},
    }, {
        'url': 'https://www.floatplane.com/post/65B5PNoBtf',
        'info_dict': {
            'id': '65B5PNoBtf',
            'description': 'I recorded the inbuilt demo mode for your 90\'s enjoyment, thanks for being Floaties!',
            'display_id': '65B5PNoBtf',
            'like_count': int,
            'release_timestamp': 1701249480,
            'uploader': 'The Trash Network',
            'availability': 'subscriber_only',
            'uploader_id': '61bc20c9a131fb692bf2a513',
            'uploader_url': 'https://www.floatplane.com/channel/TheTrashNetwork/home',
            'channel_url': 'https://www.floatplane.com/channel/TheTrashNetwork/home/thedrumthing',
            'comment_count': int,
            'title': 'The $50 electronic drum kit.',
            'channel_id': '64424fe73cd58cbcf8d8e131',
            'thumbnail': 'https://pbs.floatplane.com/blogPost_thumbnails/65B5PNoBtf/725555379422705_1701247052743.jpeg',
            'dislike_count': int,
            'channel': 'The Drum Thing',
            'release_date': '20231129',
        },
        'playlist_count': 2,
        'playlist': [{
            'info_dict': {
                'id': 'ISPJjexylS',
                'ext': 'mp4',
                'release_date': '20231129',
                'release_timestamp': 1701249480,
                'title': 'The $50 electronic drum kit. .mov',
                'channel_id': '64424fe73cd58cbcf8d8e131',
                'thumbnail': 'https://pbs.floatplane.com/video_thumbnails/ISPJjexylS/335202812134041_1701249383392.jpeg',
                'availability': 'subscriber_only',
                'uploader': 'The Trash Network',
                'duration': 622,
                'channel': 'The Drum Thing',
                'uploader_id': '61bc20c9a131fb692bf2a513',
                'channel_url': 'https://www.floatplane.com/channel/TheTrashNetwork/home/thedrumthing',
                'uploader_url': 'https://www.floatplane.com/channel/TheTrashNetwork/home',
            },
        }, {
            'info_dict': {
                'id': 'qKfxu6fEpu',
                'ext': 'aac',
                'release_date': '20231129',
                'release_timestamp': 1701249480,
                'title': 'Roland TD-7 Demo.m4a',
                'channel_id': '64424fe73cd58cbcf8d8e131',
                'availability': 'subscriber_only',
                'uploader': 'The Trash Network',
                'duration': 114,
                'channel': 'The Drum Thing',
                'uploader_id': '61bc20c9a131fb692bf2a513',
                'channel_url': 'https://www.floatplane.com/channel/TheTrashNetwork/home/thedrumthing',
                'uploader_url': 'https://www.floatplane.com/channel/TheTrashNetwork/home',
            },
        }],
        'skip': 'requires subscription: "The Trash Network"',
        'params': {'skip_download': 'm3u8'},
    }]

    def _real_initialize(self):
        if not self._get_cookies(self._BASE_URL).get('sails.sid'):
            self.raise_login_required()


class FloatplaneChannelBaseIE(InfoExtractor):
    """Subclasses must set _RESULT_IE, _BASE_URL and _PAGE_SIZE"""

    def _fetch_page(self, display_id, creator_id, channel_id, page):
        query = {
            'id': creator_id,
            'limit': self._PAGE_SIZE,
            'fetchAfter': page * self._PAGE_SIZE,
        }
        if channel_id:
            query['channel'] = channel_id
        page_data = self._download_json(
            f'{self._BASE_URL}/api/v3/content/creator', display_id,
            query=query, note=f'Downloading page {page + 1}')
        for post in page_data or []:
            yield self.url_result(
                f'{self._BASE_URL}/post/{post["id"]}',
                self._RESULT_IE, id=post['id'], title=post.get('title'),
                release_timestamp=parse_iso8601(post.get('releaseDate')))

    def _real_extract(self, url):
        creator, channel = self._match_valid_url(url).group('id', 'channel')
        display_id = join_nonempty(creator, channel, delim='/')

        creator_data = self._download_json(
            f'{self._BASE_URL}/api/v3/creator/named',
            display_id, query={'creatorURL[0]': creator})[0]

        channel_data = traverse_obj(
            creator_data, ('channels', lambda _, v: v['urlname'] == channel), get_all=False) or {}

        return self.playlist_result(OnDemandPagedList(functools.partial(
            self._fetch_page, display_id, creator_data['id'], channel_data.get('id')), self._PAGE_SIZE),
            display_id, title=channel_data.get('title') or creator_data.get('title'),
            description=channel_data.get('about') or creator_data.get('about'))


class FloatplaneChannelIE(FloatplaneChannelBaseIE):
    _VALID_URL = r'https?://(?:(?:www|beta)\.)?floatplane\.com/channel/(?P<id>[\w-]+)/home(?:/(?P<channel>[\w-]+))?'
    _BASE_URL = 'https://www.floatplane.com'
    _PAGE_SIZE = 20
    _RESULT_IE = FloatplaneIE
    _TESTS = [{
        'url': 'https://www.floatplane.com/channel/linustechtips/home/ltxexpo',
        'info_dict': {
            'id': 'linustechtips/ltxexpo',
            'title': 'LTX Expo',
            'description': 'md5:9819002f9ebe7fd7c75a3a1d38a59149',
        },
        'playlist_mincount': 51,
    }, {
        'url': 'https://www.floatplane.com/channel/ShankMods/home',
        'info_dict': {
            'id': 'ShankMods',
            'title': 'Shank Mods',
            'description': 'md5:6dff1bb07cad8e5448e04daad9be1b30',
        },
        'playlist_mincount': 14,
    }, {
        'url': 'https://beta.floatplane.com/channel/bitwit_ultra/home',
        'info_dict': {
            'id': 'bitwit_ultra',
            'title': 'Bitwit Ultra',
            'description': 'md5:1452f280bb45962976d4789200f676dd',
        },
        'playlist_mincount': 200,
    }]
